from typing import Tuple, Union

import torch
from transformers.modeling_outputs import CausalLMOutputWithPast

from ....transformers.llama.modeling_llama import LlamaForCausalLM
from ..base import BaseFoundationModelMixin
from .configuration_janus_foundation import JanusFoundationConfig


class JanusFoundationModel(BaseFoundationModelMixin, LlamaForCausalLM):
    config_class = JanusFoundationConfig

    def __init__(self, config: JanusFoundationConfig, **kwargs):
        super().__init__(config, **kwargs)

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        inputs_embeds: torch.FloatTensor = None,
        position_ids: torch.LongTensor = None,
        **kwargs,
    ) -> Union[Tuple, CausalLMOutputWithPast]:
        if position_ids is not None and position_ids.ndim == 3:
            position_ids = position_ids.squeeze(1)  # bs, 1, l -> bs, l
        if inputs_embeds is not None:
            return super().forward(
                inputs_embeds=inputs_embeds,
                position_ids=position_ids,
                **kwargs,
            )
        else:
            return super().forward(
                input_ids=input_ids,
                position_ids=position_ids,
                **kwargs,
            )



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import triton

from ....utils import logging
from ....utils.device import get_device_name
from . import envvars
from .config import load_all_configs_for
from .kernel import innermost_fn, qualified_name


logger = logging.get_logger(__name__)

CATCH_ALL_ALGO_KEY = "__CATCH_ALL__"


def algo_key_scaled(names, scales, rest_key=None):
    def key_maker(**kwargs):
        lower_names = [name.lower() for name in names]
        temp = []
        for i, name in enumerate(lower_names):
            t = name + str(kwargs[names[i]] // scales[i])
            if scales[i] != 1:
                t += f"x{scales[i]}"
            temp.append(t)
        res = "_".join(temp)

        if rest_key is not None:
            for k in rest_key:
                res += f"_{kwargs[k]}"
        return res

    return key_maker


class Pretuned(triton.KernelInterface):
    def __init__(self, fn, algo_key_maker, configs):
        self.fn = fn  # In case the outer decorator cares.
        self.kernel_name = qualified_name(fn)
        self.algo_key_maker = algo_key_maker
        self.configs = configs

        assert CATCH_ALL_ALGO_KEY in self.configs

    def run(self, *args, **kwargs):
        algo_key = self.algo_key_maker(**kwargs)
        if algo_key not in self.configs:
            if not envvars.is_untuned_warning_suppressed():
                logger.debug(
                    f"Untuned case (using algo-key [{algo_key}]) is seen when invoking "
                    f"kernel [{qualified_name(self)}], performance may suffer."
                )
            extra_kwargs = self.configs[CATCH_ALL_ALGO_KEY]
        else:
            extra_kwargs = self.configs[algo_key]
        return self.fn.run(*args, **kwargs, **extra_kwargs)


# TODO: Support using `triton.autotune` as an fallback.
def pretuned(*, algo_key=None, fallback=None):
    """Decorator to annotate a Triton kernel as pre-tuned. Hyperparameters are loaded from `PRETUNED`
    in the same folder as the kernel being defined.

    By default we look up pre-tuned hyperparameters via `kernel_name, device_name`. However, users
    are allowed to provide `algo_key` option by providing a lambda that converts arguments passed
    to kernel to a string that's used as a third level key in looking up pre-tuned hyperparameters.

    Note that ONLY named arguments (but not positional arguments) are passed to `algo_key` callback.
    """

    if algo_key is None:

        def catch_all(**kwargs):
            return CATCH_ALL_ALGO_KEY

        algo_key = catch_all

    def decorator(fn: triton.KernelInterface):
        nonlocal algo_key
        nonlocal fallback

        name = qualified_name(fn)
        configs = load_all_configs_for(innermost_fn(fn))

        if CATCH_ALL_ALGO_KEY not in configs:
            # We'd like to find a fallback hyperparameter for each `device`. This is not the same one
            # as `fallback` provided to `pretuned`. The latter is used when we're running on an untuned
            # device, while the former is just a catch-all for a specific device.
            if not envvars.is_untuned_warning_suppressed():
                logger.debug(
                    f"No pre-tuned hyperparameter for kernel [{name}], using fallback config, "
                    "performance may suffer. You may have triton version or device name mismatch. "
                    f"You have triton=={triton.__version__} and device name [{get_device_name()}]",
                )
            configs.update({CATCH_ALL_ALGO_KEY: fallback})

        assert configs[CATCH_ALL_ALGO_KEY] is not None, "No usable fallback hyperparameter for kernel {name}"
        return Pretuned(
            fn,
            algo_key_maker=algo_key,
            configs=configs,
        )

    return decorator



# adapted from https://huggingface.co/deepseek-ai/DeepSeek-V3/blob/main/inference/fp8_cast_bf16.py
import json
import os
from argparse import ArgumentParser
from glob import glob

import torch
from kernel import weight_dequant
from safetensors.torch import load_file
from tqdm import tqdm

from veomni.models import save_model_weights
from veomni.utils.device import empty_cache, get_device_type


def main(fp8_path, bf16_path):
    torch.set_default_dtype(torch.bfloat16)
    os.makedirs(bf16_path, exist_ok=True)
    model_index_file = os.path.join(fp8_path, "model.safetensors.index.json")
    with open(model_index_file) as f:
        model_index = json.load(f)
    weight_map = model_index["weight_map"]

    # Cache for loaded safetensor files
    loaded_files = {}
    fp8_weight_names = []

    # Helper function to get tensor from the correct file
    def get_tensor(tensor_name):
        file_name = weight_map[tensor_name]
        if file_name not in loaded_files:
            file_path = os.path.join(fp8_path, file_name)
            loaded_files[file_name] = load_file(file_path, device=get_device_type())
        return loaded_files[file_name][tensor_name]

    safetensor_files = list(glob(os.path.join(fp8_path, "*.safetensors")))
    safetensor_files.sort()
    new_state_dict = {}
    for safetensor_file in tqdm(safetensor_files):
        file_name = os.path.basename(safetensor_file)
        current_state_dict = load_file(safetensor_file, device=get_device_type())
        loaded_files[file_name] = current_state_dict

        for weight_name, weight in current_state_dict.items():
            if weight_name.endswith("_scale_inv"):
                continue
            elif weight.element_size() == 1:  # FP8 weight
                scale_inv_name = f"{weight_name}_scale_inv"
                try:
                    # Get scale_inv from the correct file
                    scale_inv = get_tensor(scale_inv_name)
                    fp8_weight_names.append(weight_name)
                    new_state_dict[weight_name] = weight_dequant(weight, scale_inv).cpu()
                except KeyError:
                    print(f"Warning: Missing scale_inv tensor for {weight_name}, skipping conversion")
                    new_state_dict[weight_name] = weight.cpu()
            else:
                new_state_dict[weight_name] = weight.cpu()

        # Memory management: keep only the 2 most recently used files
        if len(loaded_files) > 2:
            oldest_file = next(iter(loaded_files))
            del loaded_files[oldest_file]
            empty_cache()

    num_experts = 256
    start_layer, end_layer = 3, 61
    for i in range(start_layer, end_layer):
        gate_proj = []
        for j in range(num_experts):
            gate_proj.append(new_state_dict.pop(f"model.layers.{i}.mlp.experts.{j}.gate_proj.weight"))

        new_state_dict[f"model.layers.{i}.mlp.experts.fc1_1"] = torch.stack(gate_proj)
        up_proj = []
        for j in range(num_experts):
            up_proj.append(new_state_dict.pop(f"model.layers.{i}.mlp.experts.{j}.up_proj.weight"))

        new_state_dict[f"model.layers.{i}.mlp.experts.fc1_2"] = torch.stack(up_proj)
        down_proj = []
        for j in range(num_experts):
            down_proj.append(new_state_dict.pop(f"model.layers.{i}.mlp.experts.{j}.down_proj.weight"))

        new_state_dict[f"model.layers.{i}.mlp.experts.fc2"] = torch.stack(down_proj)

    weight_names = list(new_state_dict.keys())
    for k in weight_names:
        if f"model.layers.{end_layer}" in k:
            new_state_dict.pop(k)

    save_model_weights(bf16_path, new_state_dict)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--input-fp8-hf-path", type=str, required=True)
    parser.add_argument("--output-bf16-hf-path", type=str, required=True)
    args = parser.parse_args()
    main(args.input_fp8_hf_path, args.output_bf16_hf_path)



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional

import torch
import torch.distributed.nn.functional as distF
import torch.nn.functional as F
from torch import nn
from transformers import GenerationMixin, PreTrainedModel
from transformers.modeling_outputs import ModelOutput

from ...data.constants import IGNORE_INDEX
from ...distributed.parallel_plan import ParallelPlan
from ...distributed.parallel_state import get_parallel_state
from ...distributed.sequence_parallel import (
    gather_heads_scatter_seq,
    gather_seq_scatter_heads,
)
from ...utils import logging
from ..loader import get_model_class
from .configuration_seed_omni import SeedOmniConfig, SeedOmniDecoderConfig, SeedOmniEncoderConfig
from .decoder import BaseDecoderModelMixin, BaseDecoderOutput
from .encoder import BaseEncoderModelMixin
from .foundation import BaseFoundationModelMixin


logger = logging.get_logger(__name__)

if TYPE_CHECKING:
    from transformers import Cache


def extract_model_inputs(prefix: str, kwargs: Dict[str, "torch.Tensor"]):
    model_inputs = {}
    for key, value in kwargs.items():
        if key.startswith(prefix):
            model_inputs[key[len(prefix) :]] = value
    return model_inputs


@dataclass
class SeedOmniOutput(ModelOutput):
    logits: Optional[torch.FloatTensor] = None
    hidden_states: Optional[torch.FloatTensor] = None
    losses: Optional[Dict[str, torch.FloatTensor]] = None


class SeedOmniPreTrainedModel(PreTrainedModel):
    config_class = SeedOmniConfig
    supports_gradient_checkpointing = True
    _skip_keys_device_placement = "past_key_values"
    _supports_cache_class = True
    _supports_static_cache = True
    _supports_flash_attn_2 = True
    _supports_sdpa = True

    @property
    def _no_split_modules(self):
        no_split_modules = []
        for module in self.children():
            if isinstance(module, PreTrainedModel) and module._no_split_modules:
                no_split_modules.extend(module._no_split_modules)
            elif isinstance(module, nn.ModuleDict):
                for sub_module in module.children():
                    if isinstance(sub_module, PreTrainedModel) and sub_module._no_split_modules:
                        no_split_modules.extend(sub_module._no_split_modules)

        return no_split_modules

    @_no_split_modules.setter
    def _no_split_modules(self, value):
        pass

    def _init_weights(self, module):
        std = self.config.initializer_range
        if isinstance(module, (nn.Linear, nn.Conv3d)):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()


class SeedOmniEncoderModel(SeedOmniPreTrainedModel):
    config_class = SeedOmniEncoderConfig

    def __init__(self, config: SeedOmniEncoderConfig):
        super().__init__(config)
        torch_dtype = torch.get_default_dtype()
        self.text_encoder = nn.Embedding(
            config.text_config.vocab_size,
            config.text_config.hidden_size,
            padding_idx=config.text_config.pad_token_id,
            dtype=torch_dtype,
        )
        self.modality = []
        if config.image_config.model_type:
            model_cls = get_model_class(config.image_config)
            self.image_encoder: BaseEncoderModelMixin = model_cls._from_config(
                config.image_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
            )
            self.modality.append("image")
            self.modality.append("video")  # image encoder could be used for video embedding

        if config.video_config.model_type:
            model_cls = get_model_class(config.video_config)
            self.video_encoder: BaseEncoderModelMixin = model_cls._from_config(
                config.video_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
            )
            self.modality.append("video") if "video" not in self.modality else None

        if config.audio_config.model_type:
            model_cls = get_model_class(config.audio_config)
            self.audio_encoder: BaseEncoderModelMixin = model_cls._from_config(
                config.audio_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
            )
            self.modality.append("audio")

        self.encode_input = config.encode_input
        self.encode_output = config.encode_output

    def set_projector_trainable_only(self):
        for module in self.children():
            if isinstance(module, BaseEncoderModelMixin):
                module.set_projector_trainable_only()

    def image_forward(self, inputs_embeds: torch.Tensor, decoder_inputs, **kwargs):
        if self.encode_input:
            input_image_inputs = extract_model_inputs("image_input_", kwargs)
            input_image_mask: torch.Tensor = input_image_inputs.pop("mask", None)
            if input_image_inputs:
                input_image_features: torch.Tensor = self.image_encoder.lm_encode(**input_image_inputs).to(
                    inputs_embeds
                )
                if self.training and get_parallel_state().sp_enabled:
                    input_image_features = gather_seq_scatter_heads(
                        input_image_features, seq_dim=0, head_dim=1, group=get_parallel_state().sp_group
                    )
                input_image_features = input_image_features[: input_image_mask.sum()]
                image_mask = input_image_mask.unsqueeze(-1).expand_as(inputs_embeds).to(inputs_embeds.device)
                inputs_embeds = inputs_embeds.masked_scatter(image_mask, input_image_features)
            elif self.training:
                dummy_embeds: torch.Tensor = self.image_encoder.lm_dummy_encode()
                inputs_embeds += dummy_embeds.mean() * 0.0

        if self.encode_output:
            output_image_inputs = extract_model_inputs("image_output_", kwargs)
            output_image_mask: torch.Tensor = output_image_inputs.pop("mask", None)
            if output_image_inputs:
                output_image_features: torch.Tensor = self.image_encoder.lm_encode(**output_image_inputs).to(
                    inputs_embeds
                )
                if self.training and get_parallel_state().sp_enabled:
                    output_image_features = gather_seq_scatter_heads(
                        output_image_features, seq_dim=0, head_dim=1, group=get_parallel_state().sp_group
                    )
                output_image_features = output_image_features[: output_image_mask.sum()]
                image_mask = output_image_mask.unsqueeze(-1).expand_as(inputs_embeds).to(inputs_embeds.device)
                inputs_embeds = inputs_embeds.masked_scatter(image_mask, output_image_features)
                decoder_inputs["image_output_labels"] = output_image_features
            elif self.training:
                dummy_embeds: torch.Tensor = self.image_encoder.lm_dummy_encode()
                inputs_embeds += dummy_embeds.mean() * 0.0
        return inputs_embeds

    def video_forward(self, inputs_embeds: torch.Tensor, decoder_inputs, **kwargs):
        if self.encode_input:
            input_video_inputs = extract_model_inputs("video_input_", kwargs)
            input_video_mask: torch.Tensor = input_video_inputs.pop("mask", None)
            if input_video_inputs:
                if getattr(self, "video_encoder", None) is not None:
                    input_video_features: torch.Tensor = self.video_encoder.lm_encode(**input_video_inputs).to(
                        inputs_embeds
                    )
                else:
                    input_video_features: torch.Tensor = self.image_encoder.lm_encode(**input_video_inputs).to(
                        inputs_embeds
                    )
                if self.training and get_parallel_state().sp_enabled:
                    input_video_features = gather_seq_scatter_heads(
                        input_video_features, seq_dim=0, head_dim=1, group=get_parallel_state().sp_group
                    )
                input_video_features = input_video_features[: input_video_mask.sum()]
                video_mask = input_video_mask.unsqueeze(-1).expand_as(inputs_embeds).to(inputs_embeds.device)
                inputs_embeds = inputs_embeds.masked_scatter(video_mask, input_video_features)
            elif self.training:
                if getattr(self, "video_encoder", None) is not None:
                    dummy_embeds: torch.Tensor = self.video_encoder.lm_dummy_encode()
                else:
                    dummy_embeds: torch.Tensor = self.image_encoder.lm_dummy_encode()
                inputs_embeds += dummy_embeds.mean() * 0.0

        # TODO: video generation encode
        return inputs_embeds

    def audio_forward(self, inputs_embeds: torch.Tensor, decoder_inputs, **kwargs):
        if self.encode_input:
            input_audio_inputs = extract_model_inputs("audio_input_", kwargs)
            input_audio_mask: torch.Tensor = input_audio_inputs.pop("mask", None)
            if input_audio_inputs and input_audio_mask.sum() > 0:
                input_audio_features: torch.Tensor = self.audio_encoder.lm_encode(**input_audio_inputs).to(
                    inputs_embeds
                )
                # TODO: sp_check
                if self.training and get_parallel_state().sp_enabled:
                    input_audio_features = gather_seq_scatter_heads(
                        input_audio_features, seq_dim=0, head_dim=1, group=get_parallel_state().sp_group
                    )
                input_audio_features = input_audio_features[: input_audio_mask.sum()]
                audio_mask = input_audio_mask.unsqueeze(-1).expand_as(inputs_embeds).to(inputs_embeds.device)
                inputs_embeds = inputs_embeds.masked_scatter(audio_mask, input_audio_features)
            elif self.training:
                dummy_embeds: torch.Tensor = self.audio_encoder.lm_dummy_encode()
                inputs_embeds += dummy_embeds.mean() * 0.0

        # TODO: audio generation encode
        return inputs_embeds

    def forward(self, input_ids: torch.Tensor, **kwargs: torch.Tensor) -> Dict[str, torch.Tensor]:
        inputs_embeds: torch.Tensor = self.text_encoder(input_ids)
        decoder_inputs = {}

        if self.training and get_parallel_state().sp_enabled:
            inputs_embeds = gather_seq_scatter_heads(
                inputs_embeds, seq_dim=1, head_dim=2, group=get_parallel_state().sp_group
            )

        if "image" in self.modality:
            inputs_embeds = self.image_forward(inputs_embeds, decoder_inputs, **kwargs)

        if "video" in self.modality:
            inputs_embeds = self.video_forward(inputs_embeds, decoder_inputs, **kwargs)

        if "audio" in self.modality:
            inputs_embeds = self.audio_forward(inputs_embeds, decoder_inputs, **kwargs)

        if self.training and get_parallel_state().sp_enabled:
            inputs_embeds = gather_heads_scatter_seq(
                inputs_embeds, head_dim=2, seq_dim=1, group=get_parallel_state().sp_group
            )
        return {"inputs_embeds": inputs_embeds, "decoder_inputs": decoder_inputs}


class SeedOmniDecoderModel(SeedOmniPreTrainedModel):
    config_class = SeedOmniDecoderConfig

    def __init__(self, config: SeedOmniDecoderConfig):
        self.config = config
        super().__init__(config)
        torch_dtype = torch.get_default_dtype()
        self.modality = []
        if config.image_config.model_type:
            model_cls = get_model_class(config.image_config)
            self.image_decoder: BaseDecoderModelMixin = model_cls._from_config(
                config.image_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
            )
            self.modality.append("image")  # TODO: config keys for sp_slice
        if config.video_config.model_type:
            model_cls = get_model_class(config.video_config)
            self.video_decoder: BaseDecoderModelMixin = model_cls._from_config(
                config.video_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
            )
            self.modality.append("video")
        if config.audio_config.model_type:
            model_cls = get_model_class(config.audio_config)
            self.audio_decoder: BaseDecoderModelMixin = model_cls._from_config(
                config.audio_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
            )
            self.modality.append("audio")

        self.encode_input = config.encode_input
        self.encode_output = config.encode_output

    def set_projector_trainable_only(self):
        for module in self.children():
            if isinstance(module, BaseDecoderModelMixin):
                module.set_projector_trainable_only()

    def image_encode(self, inputs_embeds: torch.Tensor, decoder_inputs, **kwargs):
        if self.encode_input:
            input_image_inputs = extract_model_inputs("image_input_", kwargs)
            input_image_mask: torch.Tensor = input_image_inputs.pop("mask", None)
            if input_image_inputs:
                input_image_features, _ = self.image_decoder.lm_encode(**input_image_inputs)
                if self.training and get_parallel_state().sp_enabled:
                    input_image_features = gather_seq_scatter_heads(
                        input_image_features, seq_dim=0, head_dim=1, group=get_parallel_state().sp_group
                    )
                input_image_features = input_image_features[: input_image_mask.sum()]
                image_mask = input_image_mask.unsqueeze(-1).expand_as(inputs_embeds).to(inputs_embeds.device)
                image_features = input_image_features.to(inputs_embeds)
                inputs_embeds = inputs_embeds.masked_scatter(image_mask, image_features)
            elif self.training:
                dummy_embeds, _ = self.image_decoder.lm_dummy_encode()
                inputs_embeds += dummy_embeds.mean() * 0.0

        if self.encode_output:
            output_image_inputs = extract_model_inputs("image_output_", kwargs)
            output_image_mask: torch.Tensor = output_image_inputs.pop("mask", None)
            if output_image_inputs:
                output_image_features, output_image_indices = self.image_decoder.lm_encode(**output_image_inputs)
                if self.training and get_parallel_state().sp_enabled:
                    output_image_features = gather_seq_scatter_heads(
                        output_image_features, seq_dim=0, head_dim=1, group=get_parallel_state().sp_group
                    )
                output_image_features = output_image_features[: output_image_mask.sum()]
                image_mask = output_image_mask.unsqueeze(-1).expand_as(inputs_embeds).to(inputs_embeds.device)
                inputs_embeds = inputs_embeds.masked_scatter(image_mask, output_image_features)
                decoder_inputs["image_output_labels"] = output_image_indices
            elif self.training:
                dummy_embeds, _ = self.image_decoder.lm_dummy_encode()
                inputs_embeds += dummy_embeds.mean() * 0.0
        return inputs_embeds

    def encode(self, inputs_embeds: torch.Tensor, **kwargs):
        """Encodes the output images to foundation input embeds and image indices labels.
        Only support descrete tokenizer encode.
        Returns of decoder['mm_type'].encode:
            - embeds (torch.Tensor(dtype=float32, shape=(batch_size, seq_len, hidden_size)): input_embeds
            - indices (torch.Tensor(dtype=int64, shape=(batch_size, seq_len))): feature code
        """
        decoder_inputs = kwargs.pop("decoder_inputs", {})

        if self.training and get_parallel_state().sp_enabled:
            inputs_embeds = gather_seq_scatter_heads(
                inputs_embeds, seq_dim=1, head_dim=2, group=get_parallel_state().sp_group
            )

        if "image" in self.modality:
            inputs_embeds = self.image_encode(inputs_embeds, decoder_inputs, **kwargs)

        if self.training and get_parallel_state().sp_enabled:
            inputs_embeds = gather_heads_scatter_seq(
                inputs_embeds, seq_dim=1, head_dim=2, group=get_parallel_state().sp_group
            )

        return {"inputs_embeds": inputs_embeds, "decoder_inputs": decoder_inputs}

    def image_decode(self, hidden_states: torch.Tensor, decoder_inputs, loss, **kwargs):
        image_output_labels = decoder_inputs.get("image_output_labels", None)
        target_inputs = extract_model_inputs("image_target", kwargs)
        if target_inputs or image_output_labels is not None:
            output_image_inputs = extract_model_inputs("image_output_", kwargs)
            output_image_mask: torch.Tensor = output_image_inputs.pop("mask", None)
            if self.training and get_parallel_state().sp_enabled:
                bs = output_image_mask.size(0)
                sp_size = get_parallel_state().sp_size
                sp_rank = get_parallel_state().sp_rank
                sp_chunk_size = output_image_mask.size(-1) // sp_size
                sp_slice_dim = 1  # bs, seq, ...

                output_image_mask = F.pad(output_image_mask[..., 1:], (0, 1), value=0)

                labels_shape = [bs] + list(image_output_labels.shape)
                labels_shape[1] = output_image_mask.shape[1]
                labels = torch.zeros(labels_shape).to(image_output_labels)

                gathered_image_output_labels = distF.all_gather(
                    image_output_labels, group=get_parallel_state().sp_group
                )
                gathered_image_output_labels = torch.cat(gathered_image_output_labels, dim=0)
                gathered_image_output_labels = gathered_image_output_labels[: output_image_mask.sum()]

                if len(labels_shape) == 3:  # bs, seq, dim
                    output_image_mask = output_image_mask.unsqueeze(-1).expand_as(labels)
                labels[~output_image_mask] = IGNORE_INDEX
                labels = labels.masked_scatter(output_image_mask, gathered_image_output_labels)
                labels = labels.narrow(sp_slice_dim, sp_rank * sp_chunk_size, sp_chunk_size)

                outputs: BaseDecoderOutput = self.image_decoder.lm_head(
                    hidden_states,
                    labels=labels,
                    **target_inputs,
                )
                loss["image_decoder_loss"] = outputs.loss
            else:
                output_image_hs = hidden_states[..., :-1, :][output_image_mask[..., 1:]]
                labels = image_output_labels
                outputs: BaseDecoderOutput = self.image_decoder.lm_head(
                    output_image_hs,
                    labels=labels,
                    **target_inputs,
                )
                loss["image_decoder_loss"] = outputs.loss
        elif self.training:
            dummy_hidden_states = hidden_states[..., -1, :]
            outputs: BaseDecoderOutput = self.image_decoder.lm_head(dummy_hidden_states)
            loss["image_decoder_loss"] = outputs.logits.mean() * 0.0

    def decode(self, hidden_states: torch.Tensor, decoder_inputs: dict = {}, **kwargs):
        loss = {}
        if "image" in self.modality:
            self.image_decode(hidden_states, decoder_inputs, loss, **kwargs)

        return loss

    def lm_embed(self, hidden_states: torch.Tensor, model_type: str = "image", **kwargs):
        if model_type == "image":
            outputs = self.image_decoder.lm_embed(hidden_states, **kwargs)
        else:
            raise NotImplementedError
        return outputs

    def generate(self, hidden_states: torch.Tensor, modal_type: str = "image", **kwargs):
        if modal_type == "image":
            outputs = self.image_decoder.lm_generate(hidden_states, **kwargs)
        else:
            raise NotImplementedError
        return outputs


class SeedOmniModel(SeedOmniPreTrainedModel, GenerationMixin):
    def __init__(self, config: SeedOmniConfig):
        super().__init__(config)
        self.config = config
        torch_dtype = torch.get_default_dtype()
        model_cls = get_model_class(config.foundation_config)
        self.foundation: BaseFoundationModelMixin = model_cls._from_config(
            config.foundation_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
        )
        self.encoder = SeedOmniEncoderModel._from_config(
            config.encoder_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
        )
        self.decoder = SeedOmniDecoderModel._from_config(
            config.decoder_config, attn_implementation=config._attn_implementation, torch_dtype=torch_dtype
        )

    def get_input_embeddings(self):
        return self.encoder.text_encoder

    def set_input_embeddings(self, value):
        self.encoder.text_encoder = value

    def get_output_embeddings(self):
        return self.foundation.get_output_embeddings()

    def set_output_embeddings(self, new_embeddings):
        self.foundation.set_output_embeddings(new_embeddings)

    def get_parallel_plan(self):
        parallel_plan: ParallelPlan = self.foundation.get_parallel_plan()
        parallel_plan.update_prefix("foundation")
        return parallel_plan

    def resize_token_embeddings(
        self,
        new_num_tokens: Optional[int] = None,
        pad_to_multiple_of: Optional[int] = None,
        mean_resizing: bool = True,
    ):
        model_embeds = super().resize_token_embeddings(new_num_tokens, pad_to_multiple_of, mean_resizing)
        vocab_size = self.vocab_size
        self.foundation.vocab_size = vocab_size
        self.foundation.config.get_text_config().vocab_size = vocab_size
        self.config.encoder_config.text_config.vocab_size = vocab_size
        return model_embeds

    def get_modality(self):
        input_modality = self.encoder.modality
        output_modality = self.decoder.modality
        return {"input": input_modality, "output": output_modality}

    def get_position_id_func(self):
        """
        func(input_ids=input_ids, **kwargs) -> dict(position_ids=position_ids, **kwargs)
        """
        return self.foundation.position_id_func

    def forward(self, **inputs: torch.Tensor):
        decoder_inputs = {}

        if "inputs_embeds" not in inputs:
            encoder_encodes = self.encoder.forward(**inputs)
            decoder_encodes = self.decoder.encode(**inputs, **encoder_encodes)
            inputs["inputs_embeds"] = decoder_encodes["inputs_embeds"]
            decoder_inputs = decoder_encodes["decoder_inputs"]

        inputs["return_dict"] = True
        inputs["output_hidden_states"] = True
        outputs = self.foundation(**inputs)

        loss = None
        if outputs.loss is not None:
            if torch.isnan(outputs.loss):
                outputs.loss = torch.nan_to_num(outputs.loss)
            loss = outputs.loss

            hidden_states = outputs.hidden_states[-1]
            decoder_loss = self.decoder.decode(hidden_states=hidden_states, decoder_inputs=decoder_inputs, **inputs)
            losses = {"foundation_loss": loss}
            for key, v in decoder_loss.items():
                losses[key] = v

            return SeedOmniOutput(
                losses=losses,
                logits=outputs.logits,
                hidden_states=outputs.hidden_states,
            )
        return outputs

    def _prepare_image_generation_config(
        self,
        image_start_token: int = None,
        image_end_token: int = None,
        image_token_num: List = None,
        image_parallel_size: int = 16,
        image_classifier_free_guidance: bool = True,
        image_generation_config: dict = {},
        **kwargs,
    ):
        self.image_start_token = image_start_token
        self.image_token_num = image_token_num
        self.image_parallel_size = image_parallel_size
        self.image_classifier_free_guidance = image_classifier_free_guidance
        self.image_generation_config = image_generation_config
        # TODO: swift gen config based on text token (text token in image position should be trained)
        self.image_end_token = image_end_token
        return kwargs

    def _prepare_generation_config(self, *args, force_image_gen: bool = False, **kwargs):
        kwargs = self._prepare_image_generation_config(**kwargs)

        self.force_image_gen = force_image_gen
        if self.force_image_gen:
            self.parallel_size = self.image_parallel_size
            kwargs = self.setup_image_generation(**kwargs)
        else:
            kwargs = self.setup_text_generation(**kwargs)

        return super()._prepare_generation_config(*args, **kwargs)

    def setup_image_generation(self, **kwargs):
        self.gen_type = "image"
        self.tmp_image = []
        self.generated_images_sequence = []
        if hasattr(self.foundation, "get_generation_position_id"):
            # TODO: if customized rope for image generation
            self.generation_position_id_map = None
        else:  # 1d rope
            self.generation_position_id_map = None

        kwargs["input_ids"] = kwargs["input_ids"].repeat(self.image_parallel_size, 1)
        kwargs["attention_mask"] = kwargs["attention_mask"].repeat(self.image_parallel_size, 1)
        kwargs["position_ids"] = kwargs["position_ids"].repeat_interleave(self.image_parallel_size, dim=0)
        if self.image_classifier_free_guidance:
            if kwargs.get("bos_token_id", None) and kwargs["input_ids"][0][0] == kwargs["bos_token_id"]:
                start_id = 1
            else:
                start_id = 0
            if kwargs["input_ids"][0][-1] == self.image_start_token:
                end_id = -1
            else:
                end_id = 0

            kwargs["input_ids"][1::2, start_id:end_id] = kwargs.get("pad_token_id", 0)
        return kwargs

    def setup_text_generation(self, **kwargs):
        self.gen_type = "text"
        self.generation_position_id_map = None
        return kwargs

    def prepare_inputs_for_generation(
        self,
        input_ids: torch.Tensor,
        past_key_values: Optional["Cache"] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        cache_position: Optional[torch.Tensor] = None,
        **kwargs,
    ):  # only support bs=1 inference
        foundation_args = set(inspect.signature(self.foundation.prepare_inputs_for_generation).parameters)
        encoder_decoder_inputs = {}

        for key in list(kwargs.keys()):
            encoder_decoder_inputs[key] = kwargs.pop(key)
            if key in foundation_args:
                kwargs[key] = encoder_decoder_inputs.pop(key)

        model_inputs = self.foundation.prepare_inputs_for_generation(
            input_ids=input_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            cache_position=cache_position,
            **kwargs,
        )
        if cache_position[0] == 0:  # first time inference
            encoder_encodes = self.encoder(input_ids=input_ids, **encoder_decoder_inputs)
            model_inputs["inputs_embeds"] = encoder_encodes["inputs_embeds"]
            return model_inputs

        model_inputs.pop("position_ids", None)
        if self.gen_type == "text":
            if input_ids[0][-1] == self.image_start_token:
                self.setup_image_generation()
            # TODO: other modality control
        elif self.gen_type == "image":
            hidden_states = encoder_decoder_inputs["hidden_states"][-1]  # bs, cache_len, dim
            input_embeds, next_tokens = self.decoder.lm_embed(
                hidden_states[:, -1:], model_type="image", **self.image_generation_config
            )
            self.tmp_image.append(next_tokens)
            generated_tokens = len(self.tmp_image)
            model_inputs["inputs_embeds"] = input_embeds
            if self.generation_position_id_map is not None:  # TODO: customized rope for image generation
                position_ids = self.generation_position_id_map[..., generated_tokens : generated_tokens + 1]
                model_inputs["position_ids"] = position_ids + cache_position[0] - generated_tokens
            if generated_tokens == self.image_token_num:
                tmp_image = torch.cat(self.tmp_image, dim=-1)
                self.generated_images_sequence.append(tmp_image)
                self.tmp_image = []
                self.setup_text_generation()
        else:
            raise NotImplementedError
        return model_inputs

    def generate_multimodal(self, hidden_states, modal_type="image"):  # TODO: other modal_type
        return self.decoder.generate(hidden_states, modal_type=modal_type, **self.image_generation_config)

    def _validate_model_kwargs(self, model_kwargs):
        pass

    def _update_model_kwargs_for_generation(self, outputs, model_kwargs, **kwargs):
        model_kwargs = self.foundation._update_model_kwargs_for_generation(
            outputs=outputs, model_kwargs=model_kwargs, **kwargs
        )
        model_kwargs["hidden_states"] = outputs.hidden_states
        return model_kwargs

    def _has_unfinished_sequences(self, *args, **kwargs) -> bool:
        if self.gen_type != "text":
            return True
        return super()._has_unfinished_sequences(*args, **kwargs)



# Copyright 2023 Zhongjie Duan
# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

import imageio
import numpy as np
import pandas as pd
import torch
import torchvision
from einops import rearrange
from PIL import Image
from torchvision import transforms
from torchvision.transforms import v2


class TextVideoDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        base_path,
        metadata_path,
        max_num_frames=81,
        frame_interval=1,
        num_frames=81,
        height=480,
        width=832,
        is_i2v=False,
    ):
        metadata = pd.read_csv(metadata_path)
        self.path = [os.path.join(base_path, "train", file_name) for file_name in metadata["file_name"]]
        self.text = metadata["text"].to_list()

        self.max_num_frames = max_num_frames
        self.frame_interval = frame_interval
        self.num_frames = num_frames
        self.height = height
        self.width = width
        self.is_i2v = is_i2v

        self.frame_process = v2.Compose(
            [
                v2.CenterCrop(size=(height, width)),
                v2.Resize(size=(height, width), antialias=True),
                v2.ToTensor(),
                v2.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
            ]
        )

    def crop_and_resize(self, image):
        width, height = image.size
        scale = max(self.width / width, self.height / height)
        image = torchvision.transforms.functional.resize(
            image,
            (round(height * scale), round(width * scale)),
            interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
        )
        return image

    def load_frames_using_imageio(
        self, file_path, max_num_frames, start_frame_id, interval, num_frames, frame_process
    ):
        reader = imageio.get_reader(file_path)
        if (
            reader.count_frames() < max_num_frames
            or reader.count_frames() - 1 < start_frame_id + (num_frames - 1) * interval
        ):
            reader.close()
            return None

        frames = []
        first_frame = None
        for frame_id in range(num_frames):
            frame = reader.get_data(start_frame_id + frame_id * interval)
            frame = Image.fromarray(frame)
            frame = self.crop_and_resize(frame)
            if first_frame is None:
                first_frame = np.array(frame)
            frame = frame_process(frame)
            frames.append(frame)
        reader.close()
        frames = torch.stack(frames, dim=0)
        frames = rearrange(frames, "T C H W -> C T H W")

        if self.is_i2v:
            return frames, first_frame
        else:
            return frames

    def load_video(self, file_path):
        start_frame_id = torch.randint(0, self.max_num_frames - (self.num_frames - 1) * self.frame_interval, (1,))[0]
        frames = self.load_frames_using_imageio(
            file_path, self.max_num_frames, start_frame_id, self.frame_interval, self.num_frames, self.frame_process
        )
        return frames

    def is_image(self, file_path):
        file_ext_name = file_path.split(".")[-1]
        if file_ext_name.lower() in ["jpg", "jpeg", "png", "webp"]:
            return True
        return False

    def load_image(self, file_path):
        frame = Image.open(file_path).convert("RGB")
        frame = self.crop_and_resize(frame)
        frame = self.frame_process(frame)
        frame = rearrange(frame, "C H W -> C 1 H W")
        return frame

    def __getitem__(self, data_id):
        text = self.text[data_id]
        path = self.path[data_id]
        if self.is_image(path):
            if self.is_i2v:
                raise ValueError(f"{path} is not a video. I2V model doesn't support image-to-image training.")
            video = self.load_image(path)
        else:
            print(f"Loading {path}...")
            video = self.load_video(path)
            if video is None:
                print("invalid video:", path)
                return None
        if self.is_i2v:
            video, first_frame = video
            data = {"text": text, "video": video, "path": path, "first_frame": first_frame}
        else:
            data = {"text": text, "video": video, "path": path}
        return data

    def __len__(self):
        return len(self.path)


class TextImageDataset(torch.utils.data.Dataset):
    def __init__(
        self, dataset_path, steps_per_epoch=10000, height=1024, width=1024, center_crop=True, random_flip=False
    ):
        self.steps_per_epoch = steps_per_epoch
        metadata = pd.read_csv(os.path.join(dataset_path, "train/metadata.csv"))
        self.path = [os.path.join(dataset_path, "train", file_name) for file_name in metadata["file_name"]]
        self.text = metadata["text"].to_list()
        self.height = height
        self.width = width
        self.image_processor = transforms.Compose(
            [
                transforms.CenterCrop((height, width)) if center_crop else transforms.RandomCrop((height, width)),
                transforms.RandomHorizontalFlip() if random_flip else transforms.Lambda(lambda x: x),
                transforms.ToTensor(),
                transforms.Normalize([0.5], [0.5]),
            ]
        )

    def __getitem__(self, index):
        data_id = torch.randint(0, len(self.path), (1,))[0]
        data_id = (data_id + index) % len(self.path)  # For fixed seed.
        text = self.text[data_id]
        image = Image.open(self.path[data_id]).convert("RGB")
        target_height, target_width = self.height, self.width
        width, height = image.size
        scale = max(target_width / width, target_height / height)
        shape = [round(height * scale), round(width * scale)]
        image = torchvision.transforms.functional.resize(
            image, shape, interpolation=transforms.InterpolationMode.BILINEAR
        )
        image = self.image_processor(image)
        return {"text": text, "image": image}

    def __len__(self):
        return self.steps_per_epoch



import inspect
from typing import List, Optional

import numpy as np
from transformers import BatchFeature, WhisperFeatureExtractor


try:
    from transformers.audio_utils import AudioInput
except Exception:
    from transformers.tokenization_utils_base import AudioInput

from ..base import BaseEncoderProcessorMixin


class Qwen25OmniAudioModelProcessor(BaseEncoderProcessorMixin, WhisperFeatureExtractor):
    valid_kwargs = BaseEncoderProcessorMixin.valid_kwargs + list(
        inspect.signature(WhisperFeatureExtractor.__init__).parameters.keys()
    )

    def __init__(
        self,
        token_num: int = None,
        token_size: List = None,
        feature_size=80,
        sampling_rate=16000,
        hop_length=160,
        chunk_length=30,
        n_fft=400,
        padding_value=0.0,
        dither=0.0,
        return_attention_mask=False,
        **kwargs,
    ) -> None:
        BaseEncoderProcessorMixin.__init__(self, token_num=token_num, token_size=token_size, **kwargs)
        WhisperFeatureExtractor.__init__(
            self,
            feature_size=feature_size,
            sampling_rate=sampling_rate,
            hop_length=hop_length,
            chunk_length=chunk_length,
            n_fft=n_fft,
            padding_value=padding_value,
            dither=dither,
            return_attention_mask=return_attention_mask,
            **kwargs,
        )

    @property
    def model_input_names(self):
        return WhisperFeatureExtractor.model_input_names

    def process(
        self,
        audios: Optional[AudioInput] = None,
        return_tensors: str = "pt",
        **kwargs,
    ) -> BatchFeature:
        audios = [audio if audio is not None else np.zeros((0,)) for audio in audios]
        output = self.__call__(
            audios,
            return_tensors=return_tensors,
            return_attention_mask=True,
            padding="max_length",
            sampling_rate=self.sampling_rate,
        )
        features = output["input_features"]
        mask = output["attention_mask"]
        feature_lengths = mask.sum(-1)
        input_lengths = (feature_lengths - 1) // 2 + 1
        num_tokens = (input_lengths - 2) // 2 + 1
        valid_mask = num_tokens > 0
        features = features[valid_mask].permute(0, 2, 1)[mask[valid_mask].bool()]  # (len, dim)
        return BatchFeature(
            data={"features": features, "num_tokens": num_tokens, "feature_lengths": feature_lengths},
            tensor_type=return_tensors,
        )



# Copyright (c) 2023-2024 DeepSeek.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from typing import List, Optional, Tuple, Union

from transformers import LlamaConfig, PretrainedConfig


class JanusVisionConfig(PretrainedConfig):
    model_type = "janus"

    def __init__(
        self,
        width: int = 1152,
        layers: Union[Tuple[int, int, int, int], int] = 27,
        heads: int = 16,
        patch_size: int = 14,
        image_size: Union[Tuple[int, int], int] = 336,
        global_pool: str = "map",
        mlp_ratio: float = 3.7362,
        class_token: bool = False,
        num_classes: int = 0,
        use_checkpoint: bool = False,
        select_feature: str = "patch",
        select_layer: int = -2,
        pixel_mean: Optional[List[float]] = None,
        pixel_std: Optional[List[float]] = None,
        ignore_head: bool = True,
        weight_init: str = "skip",
        **kwargs,
    ):
        self.width = width
        self.layers = layers
        self.heads = heads
        self.patch_size = patch_size
        self.image_size = image_size
        self.global_pool = global_pool
        self.mlp_ratio = mlp_ratio
        self.class_token = class_token
        self.num_classes = num_classes
        self.use_checkpoint = use_checkpoint
        self.select_feature = select_feature
        self.select_layer = select_layer

        # https://github.com/deepseek-ai/Janus/blob/1daa72fa409002d40931bd7b36a9280362469ead/janus/models/clip_encoder.py#L61
        self.pixel_mean = pixel_mean
        self.pixel_std = pixel_std

        # https://github.com/deepseek-ai/Janus/blob/1daa72fa409002d40931bd7b36a9280362469ead/janus/models/siglip_vit.py#L653
        self.ignore_head = ignore_head
        self.weight_init = weight_init
        if select_layer <= 0:
            self.layers = min(self.layers, self.layers + select_layer + 1)
        else:
            self.layers = min(self.layers, select_layer)

        super().__init__(**kwargs)


class JanusGenVisionConfig(PretrainedConfig):
    model_type = "janus"

    def __init__(
        self,
        codebook_size: int = 16384,
        codebook_embed_dim: int = 8,
        codebook_l2_norm: bool = True,
        codebook_show_usage: bool = True,
        commit_loss_beta: float = 0.25,
        entropy_loss_ratio: float = 0.0,
        encoder_ch_mult: List[int] = [1, 1, 2, 2, 4],
        decoder_ch_mult: List[int] = [1, 1, 2, 2, 4],
        z_channels: int = 256,
        dropout_p: float = 0.0,
        **kwargs,
    ):
        self.codebook_size = codebook_size
        self.codebook_embed_dim = codebook_embed_dim
        self.codebook_l2_norm = codebook_l2_norm
        self.codebook_show_usage = codebook_show_usage
        self.commit_loss_beta = commit_loss_beta
        self.entropy_loss_ratio = entropy_loss_ratio
        self.encoder_ch_mult = encoder_ch_mult
        self.decoder_ch_mult = decoder_ch_mult
        self.z_channels = z_channels
        self.dropout_p = dropout_p
        super().__init__(**kwargs)


class JanusConfig(PretrainedConfig):
    model_type = "janus"

    def __init__(
        self,
        vision_config=None,
        gen_vision_config=None,
        language_config=None,
        aligner_depth=2,
        aligner_projector_type="mlp_gelu",
        gen_aligner_depth=2,
        gen_aligner_projector_type="mlp_gelu",
        gen_head_embed=2048,
        **kwargs,
    ):
        if vision_config is None:
            self.vision_config = JanusVisionConfig()
        else:
            self.vision_config = JanusVisionConfig(**vision_config)

        if gen_vision_config is None:
            self.gen_vision_config = JanusGenVisionConfig()
        else:
            self.gen_vision_config = JanusGenVisionConfig(**gen_vision_config)

        if language_config is None:
            self.language_config = LlamaConfig()
        else:
            self.language_config = LlamaConfig(**language_config)

        self.n_embed = self.language_config.hidden_size
        self.aligner_input_dim = self.vision_config.width
        self.gen_aligner_input_dim = self.gen_vision_config.codebook_embed_dim
        self.image_token_size = self.gen_vision_config.codebook_size

        self.aligner_depth = aligner_depth
        self.aligner_projector_type = aligner_projector_type
        self.gen_aligner_depth = gen_aligner_depth
        self.gen_aligner_projector_type = gen_aligner_projector_type

        self.gen_head_embed = gen_head_embed
        super().__init__(**kwargs)



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .chat_template import build_chat_template
from .data_collator import (
    CollatePipeline,
    DataCollatorWithPacking,
    DataCollatorWithPadding,
    DataCollatorWithPositionIDs,
    MakeMicroBatchCollator,
    TextSequenceShardCollator,
    UnpackDataCollator,
)
from .data_loader import build_dataloader
from .dataset import build_dataset
from .dummy_dataset import build_dummy_dataset
from .multimodal.data_collator import (
    OmniDataCollatorWithPacking,
    OmniDataCollatorWithPadding,
    OmniSequenceShardCollator,
)
from .multimodal.multimodal_chat_template import build_multimodal_chat_template


__all__ = [
    "build_chat_template",
    "build_dataloader",
    "build_dummy_dataset",
    "build_multimodal_chat_template",
    "CollatePipeline",
    "DataCollatorWithPacking",
    "DataCollatorWithPadding",
    "DataCollatorWithPositionIDs",
    "MakeMicroBatchCollator",
    "TextSequenceShardCollator",
    "UnpackDataCollator",
    "OmniDataCollatorWithPacking",
    "OmniDataCollatorWithPadding",
    "OmniSequenceShardCollator",
    "build_dataset",
]



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from typing import Callable, Dict, List, Optional

import torch
from transformers import PretrainedConfig, PreTrainedModel


class BaseFoundationConfigMixin(PretrainedConfig, ABC):
    def __init__(self, vocab_size: int = 0, hidden_size: int = 0, tie_word_embeddings: bool = None, **kwargs):
        # A Foundation model must contain `vocab_size`, `hidden_size`, and `tie_word_embeddings`
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.tie_word_embeddings = tie_word_embeddings
        super().__init__(
            vocab_size=vocab_size, hidden_size=hidden_size, tie_word_embeddings=tie_word_embeddings, **kwargs
        )


def base_position_id_func(input_ids: torch.Tensor, attention_mask: torch.Tensor, **kwargs) -> Dict[str, torch.Tensor]:
    if len(input_ids.shape) == 1:
        input_ids = input_ids.unsqueeze(0)
    if len(attention_mask.shape) == 1:
        attention_mask = attention_mask.unsqueeze(0)

    return dict(input_ids=input_ids, attention_mask=attention_mask, **kwargs)


class PositionIDFuncCompose:
    def __init__(self, customized_funcs: List[Callable]):
        self.transforms = [base_position_id_func] + customized_funcs

    def __call__(self, **x):
        for t in self.transforms:
            x = t(**x)
        return x


class BaseFoundationModelMixin(PreTrainedModel):
    def get_generation_position_id(self, **kwargs) -> Dict[str, torch.Tensor]:
        """ """
        return None

    def get_position_id_func(self) -> List[Callable]:
        """ """
        return None

    @property
    def position_id_func(self) -> Optional[Callable]:
        if self.get_position_id_func() is None:
            return None
        return PositionIDFuncCompose(self.get_position_id_func())



from functools import partial
from types import SimpleNamespace
from typing import Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from transformers.cache_utils import DynamicCache
from transformers.modeling_flash_attention_utils import FlashAttentionKwargs
from transformers.modeling_outputs import BaseModelOutputWithPast, CausalLMOutputWithPast
from transformers.processing_utils import Unpack

from .....data.constants import AUDIO_INPUT_INDEX, IGNORE_INDEX, IMAGE_INPUT_INDEX, VIDEO_INPUT_INDEX
from .....distributed.parallel_state import get_parallel_state
from .....distributed.sequence_parallel import reduce_sequence_parallel_loss
from .....utils import logging
from .....utils.import_utils import is_liger_kernel_available
from ....transformers.qwen2_5_omni.modeling_qwen2_5_omni import (
    Qwen2_5OmniThinkerForConditionalGeneration,
    apply_multimodal_rotary_pos_emb,
)
from ....transformers.qwen3_moe import modeling_qwen3_moe as qwen3_moe
from ....transformers.qwen3_moe.modeling_qwen3_moe import (
    MoeCausalLMOutputWithPast,
    MoeModelOutputWithPast,
    Qwen3MoeForCausalLM,
    Qwen3MoeModel,
    Qwen3MoePreTrainedModel,
    Qwen3MoeRMSNorm,
    Qwen3MoeRotaryEmbedding,
    load_balancing_loss_func,
)
from ....transformers.qwen3_moe.modeling_qwen3_moe import Qwen3MoeDecoderLayer as OriginalQwen3MoeDecoderLayer
from ..base import BaseFoundationModelMixin
from .configuration_qwen3_moe_foundation import Qwen3MoeFoundationConfig


if is_liger_kernel_available():
    from liger_kernel.transformers import LigerFusedLinearCrossEntropyLoss  # type: ignore
logger = logging.get_logger(__name__)


def parse_position_id_kwargs(input_ids: torch.Tensor, attention_mask: torch.Tensor, grid_thw: Dict = {}, **kwargs):
    return_dict = {
        "input_ids": input_ids,
        "attention_mask": attention_mask,
    }
    if "image" in grid_thw:
        return_dict["image_grid_thw"] = grid_thw["image"]
    if "video" in grid_thw:
        return_dict["video_grid_thw"] = grid_thw["video"]
        return_dict["second_per_grids"] = torch.tensor([1.0] * len(grid_thw["video"])).to(input_ids.device)
    if "feature_lengths" in kwargs:  # audio 'grid_thw'
        return_dict["audio_seqlens"] = kwargs["feature_lengths"]["audio"]
    return return_dict


def prepare_fa2_from_position_ids(position_ids):
    position_ids = position_ids.flatten()
    indices_q = torch.arange(position_ids.size(0), device=position_ids.device, dtype=torch.int32)

    cu_seq_lens = torch.cat(
        (
            indices_q[position_ids == 0],
            torch.tensor(position_ids.size(), device=position_ids.device, dtype=torch.int32),
        )
    )

    max_length = cu_seq_lens.diff().max()

    return (indices_q, (cu_seq_lens, cu_seq_lens), (max_length, max_length))


class Qwen3MoeFoundationRotaryEmbedding(Qwen3MoeRotaryEmbedding):
    @torch.no_grad()
    def forward(self, x, position_ids):
        if "dynamic" in self.rope_type:
            self._dynamic_frequency_update(position_ids, device=x.device)

        # Core RoPE block
        inv_freq_expanded = self.inv_freq[None, None, :, None].float().expand(3, position_ids.shape[1], -1, 1)
        position_ids_expanded = position_ids[:, :, None, :].float()
        # Force float32 (see https://github.com/huggingface/transformers/pull/29285)
        device_type = x.device.type
        device_type = device_type if isinstance(device_type, str) and device_type != "mps" else "cpu"
        with torch.autocast(device_type=device_type, enabled=False):
            freqs = (inv_freq_expanded.float().to(x.device) @ position_ids_expanded.float()).transpose(2, 3)
            emb = torch.cat((freqs, freqs), dim=-1)
            cos = emb.cos() * self.attention_scaling
            sin = emb.sin() * self.attention_scaling
        return cos.to(dtype=x.dtype), sin.to(dtype=x.dtype)


class Qwen3MoeDecoderLayer(OriginalQwen3MoeDecoderLayer):
    def __init__(self, config: Qwen3MoeFoundationConfig, layer_idx: int):
        super().__init__(config, layer_idx)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_value: Optional[Tuple[torch.Tensor]] = None,
        output_attentions: Optional[bool] = False,
        output_router_logits: Optional[bool] = False,
        use_cache: Optional[bool] = False,
        cache_position: Optional[torch.LongTensor] = None,
        position_embeddings: Optional[Tuple[torch.Tensor, torch.Tensor]] = None,  # necessary, but kept here for BC
        **kwargs: Unpack[FlashAttentionKwargs],
    ) -> Tuple[torch.FloatTensor, Optional[Tuple[torch.FloatTensor, torch.FloatTensor]]]:
        residual = hidden_states
        hidden_states = self.input_layernorm(hidden_states)
        # Self Attention
        indices_q, cu_seq_lens, max_seq_lens = prepare_fa2_from_position_ids(
            position_ids[0]
        )  # remove multimodal channel dimension
        cu_seq_lens_q, cu_seq_lens_k = cu_seq_lens
        max_length_q, max_length_k = max_seq_lens
        hidden_states, self_attn_weights = self.self_attn(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_value=past_key_value,
            output_attentions=output_attentions,
            use_cache=use_cache,
            cache_position=cache_position,
            position_embeddings=position_embeddings,
            cu_seq_lens_q=cu_seq_lens_q,
            cu_seq_lens_k=cu_seq_lens_k,
            max_length_q=max_length_q,
            max_length_k=max_length_k,
        )
        hidden_states = residual + hidden_states
        # Fully Connected
        residual = hidden_states
        hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)
        if isinstance(hidden_states, tuple):
            hidden_states, router_logits = hidden_states
        else:
            router_logits = None

        hidden_states = residual + hidden_states

        outputs = (hidden_states,)

        if output_attentions:
            outputs += (self_attn_weights,)

        if output_router_logits:
            outputs += (router_logits,)

        return outputs


class MultimodalQwen3MoeModel(Qwen3MoeModel):
    def __init__(self, config):
        Qwen3MoePreTrainedModel.__init__(self, config)
        self.padding_idx = config.pad_token_id
        self.vocab_size = config.vocab_size

        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size, self.padding_idx)
        self.layers = nn.ModuleList(
            [
                Qwen3MoeDecoderLayer(config, layer_idx) for layer_idx in range(config.num_hidden_layers)
            ]  # overwrite decoder forward
        )
        self.norm = Qwen3MoeRMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        if config.rope_type == "3d_rope":
            self.rotary_emb = Qwen3MoeFoundationRotaryEmbedding(config=config)  # overwrite position embedding
        else:
            self.rotary_emb = Qwen3MoeRotaryEmbedding(config=config)
        self.gradient_checkpointing = False

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_router_logits: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        **flash_attn_kwargs: Unpack[FlashAttentionKwargs],
    ) -> Union[Tuple, BaseModelOutputWithPast]:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_router_logits = (
            output_router_logits if output_router_logits is not None else self.config.output_router_logits
        )
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if (input_ids is None) ^ (inputs_embeds is not None):
            raise ValueError("You must specify exactly one of input_ids or inputs_embeds")

        if self.gradient_checkpointing and self.training:
            if use_cache:
                logger.warning_once(
                    "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                )
                use_cache = False

        if use_cache and past_key_values is None:
            past_key_values = DynamicCache()

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        if cache_position is None:
            past_seen_tokens = past_key_values.get_seq_length() if past_key_values is not None else 0
            cache_position = torch.arange(
                past_seen_tokens, past_seen_tokens + inputs_embeds.shape[1], device=inputs_embeds.device
            )
        if position_ids is None:
            position_ids = cache_position.view(1, 1, -1).expand(3, inputs_embeds.shape[0], -1)

        causal_mask = self._update_causal_mask(
            attention_mask, inputs_embeds, cache_position, past_key_values, output_attentions
        )

        hidden_states = inputs_embeds

        # create position embeddings to be shared across the decoder layers
        if self.config.rope_type == "3d_rope":
            position_embeddings = self.rotary_emb(hidden_states, position_ids)
        else:
            position_embeddings = self.rotary_emb(hidden_states, position_ids[0])  # 1, bs, len -> bs, len
        # decoder layers
        all_hidden_states = () if output_hidden_states else None
        all_self_attns = () if output_attentions else None
        all_router_logits = () if output_router_logits else None
        for decoder_layer in self.layers:
            if output_hidden_states:
                all_hidden_states += (hidden_states,)

            if self.gradient_checkpointing and self.training:
                layer_outputs = self._gradient_checkpointing_func(
                    decoder_layer.__call__,
                    hidden_states,
                    causal_mask,
                    position_ids,
                    past_key_values,
                    output_attentions,
                    output_router_logits,
                    use_cache,
                    cache_position,
                    position_embeddings,
                )
            else:
                layer_outputs = decoder_layer(
                    hidden_states,
                    attention_mask=causal_mask,
                    position_ids=position_ids,
                    past_key_value=past_key_values,
                    output_attentions=output_attentions,
                    output_router_logits=output_router_logits,
                    use_cache=use_cache,
                    cache_position=cache_position,
                    position_embeddings=position_embeddings,
                    **flash_attn_kwargs,
                )

            hidden_states = layer_outputs[0]

            if output_attentions:
                all_self_attns += (layer_outputs[1],)

            if output_router_logits:
                all_router_logits += (layer_outputs[-1],)

        hidden_states = self.norm(hidden_states)

        # add hidden states from the last decoder layer
        if output_hidden_states:
            all_hidden_states += (hidden_states,)

        output = MoeModelOutputWithPast(
            last_hidden_state=hidden_states,
            past_key_values=past_key_values,
            hidden_states=all_hidden_states,
            attentions=all_self_attns,
            router_logits=all_router_logits,
        )
        return output if return_dict else output.to_tuple()


class Qwen3MoeFoundationModel(BaseFoundationModelMixin, Qwen3MoeForCausalLM):
    config_class = Qwen3MoeFoundationConfig

    def __init__(self, config: Qwen3MoeFoundationConfig, **kwargs):
        Qwen3MoePreTrainedModel.__init__(self, config, **kwargs)
        self.spatial_merge_size = 2  # following qwen2vl
        self.rope_deltas = None  # for inference
        self.rope_type = config.rope_type  # 1d_rope/3d_rope

        self.model = MultimodalQwen3MoeModel(config)  # overwrite model.forward()
        self.vocab_size = config.vocab_size
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.router_aux_loss_coef = config.router_aux_loss_coef
        self.num_experts = config.num_experts
        self.num_experts_per_tok = config.num_experts_per_tok

        self.image_token_index = IMAGE_INPUT_INDEX
        self.video_token_index = VIDEO_INPUT_INDEX
        self.audio_token_index = AUDIO_INPUT_INDEX

        if self.rope_type == "3d_rope":
            self.mrope_section = [16, 24, 24]  # following qwen25omni

            qwen3_moe.apply_rotary_pos_emb = partial(apply_multimodal_rotary_pos_emb, mrope_section=self.mrope_section)

        self.post_init()

    def get_position_id_func(self):
        if self.rope_type == "3d_rope":
            fake_model = SimpleNamespace(
                config=SimpleNamespace(
                    vision_start_token_id=151669,  # only support Qwen3moe tokenizer TODO: unify this
                    audio_start_token_id=151671,
                    position_id_per_seconds=25,  # following qwen25omni
                    seconds_per_chunk=2.0,  # following qwen25omni
                ),
                image_token_index=self.image_token_index,
                video_token_index=self.video_token_index,
                audio_token_index=self.audio_token_index,
                spatial_merge_size=self.spatial_merge_size,
            )
            return [
                parse_position_id_kwargs,
                Qwen2_5OmniThinkerForConditionalGeneration.get_position_id_func(fake_model),
            ]
        else:
            return None  # the default position_id_func is torch.arange(len(input_ids))

    def prepare_inputs_for_generation(
        self,
        input_ids,
        rope_deltas=None,
        **kwargs,
    ):
        if rope_deltas is not None:
            self.rope_deltas = rope_deltas
        # TODO: support qwen25omni inference
        return super().prepare_inputs_for_generation(
            input_ids,
            **kwargs,
        )

    def forward(
        self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_router_logits: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        logits_to_keep: Union[int, torch.Tensor] = 0,
        **kwargs,
    ) -> Union[Tuple, CausalLMOutputWithPast]:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_router_logits = (
            output_router_logits if output_router_logits is not None else self.config.output_router_logits
        )

        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if not get_parallel_state().sp_enabled and labels is not None:
            # Shift so that tokens < n predict n
            labels = labels[..., 1:].contiguous()
            labels = labels.view(-1)
            if (
                position_ids is not None
                and position_ids.size(0) == 1
                and not (torch.diff(position_ids, dim=-1) >= 0).all()
            ):
                position_ids_ = position_ids[:, 0].flatten()  # multimodal position ids is (bs, dim, seq_len)
                indices_q = torch.arange(position_ids_.size(0), device=position_ids_.device, dtype=torch.int32)
                cu_seq_lens = torch.cat(
                    (
                        indices_q[position_ids_ == 0],
                        torch.tensor(position_ids_.size(), device=position_ids_.device, dtype=torch.int32),
                    )
                )
                labels[cu_seq_lens[1:-1] - 1] = IGNORE_INDEX

        if position_ids is not None:
            position_ids = position_ids.transpose(0, 1).contiguous()  # bs, dim, len -> dim, bs, len

        if inputs_embeds is not None:
            input_ids = None

        # decoder outputs consists of (dec_features, layer_state, dec_hidden, dec_attn)
        outputs = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            output_router_logits=output_router_logits,
            return_dict=return_dict,
            cache_position=cache_position,
            **kwargs,
        )

        hidden_states = outputs[0]
        # Only compute necessary logits, and do not upcast them to float if we are not computing the loss
        slice_indices = slice(-logits_to_keep, None) if isinstance(logits_to_keep, int) else logits_to_keep
        hidden_states = hidden_states[:, slice_indices, :]

        loss = None
        logits = None
        if labels is not None:
            labels = labels.view(-1)  # flatten label
            if is_liger_kernel_available():
                loss_fct = LigerFusedLinearCrossEntropyLoss(reduction="mean")
                if not get_parallel_state().sp_enabled:
                    # Shift so that tokens < n predict n
                    hidden_states = hidden_states[..., :-1, :].contiguous()

                hidden_states = hidden_states.view(-1, self.config.hidden_size)
                loss = loss_fct(self.lm_head.weight, hidden_states, labels)
            else:
                loss_fct = torch.nn.CrossEntropyLoss(reduction="mean")
                logits = self.lm_head(hidden_states)
                # Upcast to float if we need to compute the loss to avoid potential precision issues
                logits = logits.float()

                if not get_parallel_state().sp_enabled:
                    # Shift so that tokens < n predict n
                    logits = logits[..., :-1, :].contiguous()

                # Flatten the tokens
                logits = logits.view(-1, self.vocab_size)
                torch.distributed.barrier()
                loss = loss_fct(logits, labels)

            if get_parallel_state().sp_enabled:
                num_valid_tokens = (labels != IGNORE_INDEX).sum()
                loss = reduce_sequence_parallel_loss(loss, num_valid_tokens)
        else:
            logits = self.lm_head(hidden_states)

        aux_loss = None
        if output_router_logits:
            aux_loss = load_balancing_loss_func(
                outputs.router_logits if return_dict else outputs[-1],
                self.num_experts,
                self.num_experts_per_tok,
                attention_mask,
            )
            if labels is not None:
                loss += self.router_aux_loss_coef * aux_loss.to(loss.device)  # make sure to reside in the same device

        if not return_dict:
            output = (logits,) + outputs[1:]
            if output_router_logits:
                output = (aux_loss,) + output
            return (loss,) + output if loss is not None else output

        return MoeCausalLMOutputWithPast(
            loss=loss,
            aux_loss=aux_loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
            router_logits=outputs.router_logits,
        )



import math
from typing import List

import torch
import torch.distributed as dist
from torch.distributed._tensor import DTensor

from ...utils.device import IS_NPU_AVAILABLE, get_device_type
from ...utils.logging import get_logger
from ..parallel_state import get_parallel_state


logger = get_logger(__name__)


def clip_grad_norm(
    model, max_norm: float, norm_type: float = 2.0, error_if_nonfinite: bool = False, foreach: bool | None = None
) -> torch.Tensor:
    # EP-aware path (FSDP2 + EP): maintain mathematical parity with FSDP1 clipper
    if hasattr(model, "_ep_param_groups"):
        return ep_fsdp2_clip_grad_norm(
            model,
            max_norm,
            norm_type=norm_type,
            error_if_nonfinite=error_if_nonfinite,
            foreach=foreach,
        )

    grad_norm = torch.nn.utils.clip_grad_norm_(
        model.parameters(),
        max_norm,
        norm_type=norm_type,
        error_if_nonfinite=error_if_nonfinite,
        foreach=foreach,
    )
    if isinstance(grad_norm, DTensor):
        grad_norm = grad_norm.full_tensor()
    return grad_norm


@torch.no_grad()
def ep_fsdp2_clip_grad_norm(
    model, max_norm: float, norm_type: float = 2.0, error_if_nonfinite: bool = False, foreach: bool | None = None
) -> torch.Tensor:
    """
    EP-aware gradient clipping for composable FSDP2 with reductions mirroring FSDP1:

    - Compute local norms for non-EP and EP parameter groups separately.
    - For finite p: sum p-th powers across the appropriate groups, then take 1/p.
      • non-EP: all-reduce over FSDP group.
      • EP: all-reduce over EP-FSDP group, then over EP group.
    - For inf-norm: take elementwise MAX with the same reduction groups (MAX).
    - Use a single global clip coefficient for both groups.
    """

    ps = get_parallel_state()
    fsdp_group = ps.fsdp_group
    ep_group = ps.ep_group if ps.ep_enabled else None
    # For EP params sharded by FSDP2 along hidden dimension
    ep_fsdp_group = None
    if ps.ep_enabled and ps.ep_fsdp_device_mesh is not None:
        ep_fsdp_group = ps.ep_fsdp_device_mesh["ep_fsdp"].get_group()

    # Build param groups (filter out params without grads)
    ep_params: List[torch.nn.Parameter] = [p for p in model._ep_param_groups.get("ep", []) if p.grad is not None]
    non_ep_params: List[torch.nn.Parameter] = [
        p for p in model._ep_param_groups.get("non_ep", []) if p.grad is not None
    ]
    if IS_NPU_AVAILABLE and ps.ep_enabled and ep_params:
        # TODO(https://github.com/ByteDance-Seed/VeOmni/issues/241):
        # This is workaround for NPU. Need to remove this after PreSumMul is supported after NPU
        # Averaging gradients for EP params through dividing grads by ep_size
        # this is to simulate fsdp2 set_gradient_divide_factor
        scale = 1.0 / float(ps.ep_size)
        with torch.no_grad():
            for q in ep_params:
                if q.grad is not None:
                    q.grad.mul_(scale)
    # Compute and reduce non-EP
    non_ep_total = _fsdp2_reduce_group(
        params=non_ep_params,
        norm_type=norm_type,
        reduce_groups=[("fsdp", fsdp_group)],
    )
    logger.debug_rank0(f"non_ep total grad norm: {non_ep_total}")
    logger.debug_rank0(f"ep_params reduces groups: {ep_fsdp_group=}, {ep_group=}")
    # Compute and reduce EP: first across ep_fsdp, then across ep
    ep_total = _fsdp2_reduce_group(
        params=ep_params,
        norm_type=norm_type,
        reduce_groups=[("ep_fsdp", ep_fsdp_group), ("ep", ep_group)],
    )
    logger.debug_rank0(f"ep total grad norm: {ep_total}")

    if math.isinf(norm_type):
        total_norm = torch.maximum(non_ep_total, ep_total)
    else:
        total_norm = (non_ep_total + ep_total) ** (1.0 / float(norm_type))

    # Apply the same clip coefficient to both groups
    torch.nn.utils.clip_grads_with_norm_(ep_params, max_norm, total_norm, foreach=foreach)
    torch.nn.utils.clip_grads_with_norm_(non_ep_params, max_norm, total_norm, foreach=foreach)

    return total_norm


# compute local sum of param gard norm
def _local_pth_sum(params: List[torch.nn.Parameter], p: float) -> torch.Tensor:
    dev = None
    acc = None
    for q in params:
        g = q.grad
        if g is None:
            continue
        if isinstance(g, DTensor):
            g_local = g.to_local()
        else:
            g_local = g
        if dev is None:
            dev = g_local.device
            acc = torch.tensor(0.0, device=dev, dtype=torch.float32)
        # compute in FP32 for stability
        gn = torch.norm(g_local.detach().to(torch.float32), p=p)
        acc = acc + (gn**p)
    if acc is None:
        # no grads; choose a reasonable device
        dev = torch.device(get_device_type())
        acc = torch.tensor(0.0, device=dev, dtype=torch.float32)
    return acc


def _local_max(params: List[torch.nn.Parameter]) -> torch.Tensor:
    dev = None
    mx = None
    for q in params:
        g = q.grad
        if g is None:
            continue
        if isinstance(g, DTensor):
            g_local = g.to_local()
        else:
            g_local = g
        if dev is None:
            dev = g_local.device
            mx = torch.tensor(0.0, device=dev, dtype=torch.float32)
        gn = torch.max(torch.abs(g_local.detach().to(torch.float32)))
        mx = torch.maximum(mx, gn)
    if mx is None:
        dev = torch.device(get_device_type())
        mx = torch.tensor(0.0, device=dev, dtype=torch.float32)
    return mx


def _fsdp2_reduce_group(
    params: List[torch.nn.Parameter],
    norm_type: float,
    reduce_groups: List[tuple[str, dist.ProcessGroup | None]],
) -> torch.Tensor:
    """Compute local group statistic and reduce over provided groups.

    For finite p, returns the globally-reduced sum of p-th powers (not the final norm).
    For inf, returns the globally-reduced max.
    """
    if math.isinf(norm_type):
        val = _local_max(params)
        for _, group in reduce_groups:
            if group is not None:
                dist.all_reduce(val, op=dist.ReduceOp.MAX, group=group)
        return val
    else:
        p = float(norm_type)
        val = _local_pth_sum(params, p)
        logger.debug_rank0(f"local total grad norm: {val}. ProcessGroups to sum {reduce_groups}")
        for name, group in reduce_groups:
            if group is not None:
                dist.all_reduce(val, op=dist.ReduceOp.SUM, group=group)
                logger.debug_rank0(f"After Sum of group {name} total grad norm is {val}")
        return val



import math
from typing import Optional

import numpy as np
import torch
import torch.nn.functional as F
from diffusers.image_processor import VaeImageProcessor
from transformers import BaseImageProcessor, BatchFeature
from transformers.image_transforms import resize
from transformers.image_utils import (
    ImageInput,
    get_image_size,
    make_list_of_images,
    to_numpy_array,
)


def smart_resize(
    height: int, width: int, factor: int = 28, min_pixels: int = 56 * 56, max_pixels: int = 14 * 14 * 4 * 1280
):
    """
        Copied from qwen2vl.imageprocessor.smart_resize.
        Rescales the image so that the following conditions are met:

    1. Both dimensions (height and width) are divisible by 'factor'.

    2. The total number of pixels is within the range ['min_pixels', 'max_pixels'].

    3. The aspect ratio of the image is maintained as closely as possible.

    """
    if height < factor or width < factor:
        raise ValueError(f"height:{height} or width:{width} must be larger than factor:{factor}")
    elif max(height, width) / min(height, width) > 200:
        raise ValueError(
            f"absolute aspect ratio must be smaller than 200, got {max(height, width) / min(height, width)}"
        )
    h_bar = round(height / factor) * factor
    w_bar = round(width / factor) * factor
    if h_bar * w_bar > max_pixels:
        beta = math.sqrt((height * width) / max_pixels)
        h_bar = math.floor(height / beta / factor) * factor
        w_bar = math.floor(width / beta / factor) * factor
    elif h_bar * w_bar < min_pixels:
        beta = math.sqrt(min_pixels / (height * width))
        h_bar = math.ceil(height * beta / factor) * factor
        w_bar = math.ceil(width * beta / factor) * factor
    return h_bar, w_bar


class UltraEditProcessor(BaseImageProcessor):
    model_input_names = ["images"]

    def __init__(self, vae_scale_factor: int = 8, **kwargs):
        super().__init__(**kwargs)
        self.vae_scale_factor = vae_scale_factor
        self.max_pixels = 512
        self.patch_size = 14
        self.merge_size = 2
        self.image_processor = VaeImageProcessor(vae_scale_factor=vae_scale_factor)

    def to_dict(self):
        encoder_dict = super().to_dict()
        encoder_dict.pop("image_processor")
        return encoder_dict

    def transform(self, image: ImageInput) -> np.array:
        height, width = get_image_size(image)
        resized_height, resized_width = smart_resize(
            height,
            width,
            factor=self.patch_size * self.merge_size,
            max_pixels=self.max_pixels * self.max_pixels,
        )
        image = resize(image, size=(resized_height, resized_width))
        image = self.image_processor.preprocess(image)
        return image

    def preprocess(
        self,
        images: Optional[ImageInput] = None,
        return_tensors: str = "pt",
    ) -> torch.Tensor:
        images = make_list_of_images(images)
        images = [to_numpy_array(image) for image in images]

        pixel_values, image_mask, image_grid_thw = [], [], []

        for image in images:
            pixel_value = self.transform(image)
            image_grid_thw.append(
                torch.tensor([1, pixel_value.shape[2] // self.patch_size, pixel_value.shape[3] // self.patch_size])
            )
            pad_height = (self.max_pixels - pixel_value.shape[2]) // 2
            pad_width = (self.max_pixels - pixel_value.shape[3]) // 2
            mask_img = torch.ones_like(pixel_value)
            padding = (pad_width, pad_width, pad_height, pad_height)
            pixel_value = F.pad(pixel_value, padding, mode="constant", value=0)
            mask_img = F.pad(mask_img, padding, mode="constant", value=0)
            pixel_values.append(pixel_value)
            image_mask.append(mask_img)

        pixel_values = torch.cat(pixel_values, dim=0)
        image_mask = torch.cat(image_mask, dim=0)
        image_grid_thw = torch.stack(image_grid_thw, dim=0)
        return BatchFeature(
            data={"features": pixel_values, "mask": image_mask, "image_grid_thw": image_grid_thw},
            tensor_type=return_tensors,
        )



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union

from ..utils.logging import get_logger
from ..utils.registry import Registry


logger = get_logger(__name__)


CHECKPOINTER_REGISTRY = Registry("checkpointer")
CHECKPOINT_TO_STATE_DICT_REGISTRY = Registry("checkpoint_to_state_dict")


def build_checkpointer(ckpt_manager: str, dist_backend: str):
    return CHECKPOINTER_REGISTRY[ckpt_manager](dist_backend)


def ckpt_to_state_dict(
    save_checkpoint_path: Union[str, os.PathLike],
    ckpt_manager: str = "dcp",
) -> Dict[str, Any]:
    """
    Interface to convert a checkpoint to a state_dict.
    Supported checkpoint managers:
        - dcp

    Args:
        save_checkpoint_path: Path to the checkpoint.
        ckpt_manager: Checkpoint manager.
    Returns:
        state_dict: State dict.
    """
    return CHECKPOINT_TO_STATE_DICT_REGISTRY[ckpt_manager](save_checkpoint_path)


class CheckpointerBase(ABC):
    """Base class for checkpointer"""

    @abstractmethod
    def save(
        cls,
        path: str,
        state: Dict[str, Any],
        save_async: Optional[bool],
        global_steps: Optional[int],
    ):
        return

    @abstractmethod
    def load(
        cls,
        path: str,
        state: Dict[str, Any],
    ):
        return


@CHECKPOINTER_REGISTRY.register("dcp")
def dcp_checkpointer(dist_backend: str):
    from ..utils.import_utils import is_torch_version_greater_than

    if not is_torch_version_greater_than("2.4"):
        raise ValueError("DCP checkpoint manager requires torch version >= 2.4")
    if dist_backend not in ["ddp", "fsdp1", "fsdp2"]:
        raise ValueError(
            f"Unsupported distributed backend: {dist_backend} for DCP checkpoint manager, supported modes are: ddp, fsdp1, fsdp2"
        )
    from .dcp_checkpointer import DistributedCheckpointer

    return DistributedCheckpointer


@CHECKPOINT_TO_STATE_DICT_REGISTRY.register("dcp")
def dcp_ckpt_to_state_dict(save_checkpoint_path: Union[str, os.PathLike], **kwargs):
    from ..utils.import_utils import is_torch_version_greater_than

    if not is_torch_version_greater_than("2.4"):
        raise ValueError("DCP checkpoint manager requires torch version >= 2.4")
    from .dcp_checkpointer import dcp_to_torch_state_dict

    return dcp_to_torch_state_dict(save_checkpoint_path)



import os
import time
from dataclasses import dataclass, field
from typing import Any, Dict, List

import torch
import torch.distributed as dist
import torch.nn.functional as F
import wandb
from tqdm import trange

from veomni.checkpoint import build_checkpointer, ckpt_to_state_dict
from veomni.data.diffusion.data_loader import build_dit_dataloader
from veomni.data.diffusion.dataset import build_tensor_dataset
from veomni.distributed.clip_grad_norm import veomni_clip_grad_norm
from veomni.distributed.offloading import build_activation_offloading_context
from veomni.distributed.parallel_state import get_parallel_state, init_parallel_state
from veomni.distributed.torch_parallelize import build_parallelize_model
from veomni.models import (
    build_foundation_model,
    save_model_assets,
)
from veomni.optim import build_lr_scheduler, build_optimizer
from veomni.schedulers.flow_match import FlowMatchScheduler
from veomni.utils import helper
from veomni.utils.arguments import (
    DataArguments,
    ModelArguments,
    TrainingArguments,
    parse_args,
    save_args,
)
from veomni.utils.device import (
    get_device_type,
    get_dist_comm_backend,
    get_torch_device,
    synchronize,
)
from veomni.utils.dist_utils import all_reduce
from veomni.utils.dit_utils import EnvironMeter, save_model_weights
from veomni.utils.lora_utils import add_lora_to_model, freeze_parameters
from veomni.utils.recompute_utils import convert_ops_to_objects


logger = helper.create_logger(__name__)


@dataclass
class MyDataArguments(DataArguments):
    datasets_repeat: int = field(
        default=1,
        metadata={"help": "The number of times to repeat the datasets."},
    )


@dataclass
class MyTrainingArguments(TrainingArguments):
    save_initial_model: bool = field(
        default=False,
        metadata={"help": "Whether or not to save the initial model."},
    )
    ops_to_save: List[str] = field(
        default_factory=list,
        metadata={"help": "Ops to save."},
    )


@dataclass
class Arguments:
    model: ModelArguments = field(default_factory=ModelArguments)
    data: MyDataArguments = field(default_factory=MyDataArguments)
    train: MyTrainingArguments = field(default_factory=MyTrainingArguments)


def get_param_groups(model: torch.nn.Module, default_lr: float, vit_lr: float):
    vit_params, other_params = [], []
    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue
        if "visual" in name:
            vit_params.append(param)
        else:
            other_params.append(param)
    return [
        {"params": vit_params, "lr": vit_lr},
        {"params": other_params, "lr": default_lr},
    ]


def main():
    args = parse_args(Arguments)
    get_torch_device().set_device(f"{get_device_type()}:{args.train.local_rank}")
    dist.init_process_group(backend=get_dist_comm_backend())
    helper.set_seed(args.train.seed, args.train.enable_full_determinism)
    if args.train.global_rank == 0:
        save_args(args, args.train.output_dir)

    Checkpointer = build_checkpointer(
        dist_backend=args.train.data_parallel_mode,
        ckpt_manager=args.train.ckpt_manager,
    )

    init_parallel_state(
        dp_size=args.train.data_parallel_size,
        dp_replicate_size=args.train.data_parallel_replicate_size,
        dp_shard_size=args.train.data_parallel_shard_size,
        tp_size=args.train.tensor_parallel_size,
        ep_size=args.train.expert_parallel_size,
        pp_size=args.train.pipeline_parallel_size,
        cp_size=args.train.context_parallel_size,
        ulysses_size=args.train.ulysses_parallel_size,
        dp_mode=args.train.data_parallel_mode,
    )
    logger.info_rank0(
        f"Parallel state: dp:{args.train.data_parallel_mode}, tp:{args.train.tensor_parallel_size}, ep:{args.train.expert_parallel_size}, pp:{args.train.pipeline_parallel_size}, cp:{args.train.context_parallel_size}, ulysses:{args.train.ulysses_parallel_size}"
    )

    if args.data.data_type == "diffusion":
        train_dataset = build_tensor_dataset(
            base_path=args.data.train_path,
            metadata_path=os.path.join(args.data.train_path, "metadata.csv"),
            datasets_repeat=args.data.datasets_repeat,
        )
        args.train.compute_train_steps(
            args.data.max_seq_len,
            args.data.train_size,
            len(train_dataset) // args.train.data_parallel_size,
        )

        train_dataloader = build_dit_dataloader(
            dataset=train_dataset,
            micro_batch_size=args.train.micro_batch_size,
            global_batch_size=args.train.global_batch_size,
            dataloader_batch_size=args.train.dataloader_batch_size,
            seed=args.train.seed,
            train_steps=args.train.train_steps,
            num_workers=args.data.num_workers,
            drop_last=args.data.drop_last,
            pin_memory=args.data.pin_memory,
            prefetch_factor=args.data.prefetch_factor,
        )
    else:
        raise NotImplementedError(f"Unsupported data type: {args.data.data_type}.")

    model = build_foundation_model(
        config_path=args.model.config_path,
        weights_path=args.model.model_path,
        init_device=args.train.init_device,
        torch_dtype="bfloat16",
        attn_implementation=args.model.attn_implementation,
        force_use_huggingface=args.model.force_use_huggingface,
    )
    model.micro_batch_size = args.train.micro_batch_size

    model_config = model.config
    helper.print_device_mem_info("VRAM usage after building model")

    lora_target_modules_support = ["q", "k", "v", "o", "ffn.0", "ffn.2"]
    if args.train.train_architecture == "lora":
        logger.info_rank0("train_architecture is lora")
        _use_orig_params = True
        freeze_parameters(model)
        add_lora_to_model(
            model,
            lora_rank=args.model.lora_rank,
            lora_alpha=args.model.lora_alpha,
            lora_target_modules=args.model.lora_target_modules,
            init_lora_weights=args.model.init_lora_weights,
            pretrained_lora_path=args.model.pretrained_lora_path,
            lora_target_modules_support=lora_target_modules_support,
        )
        model.to(torch.bfloat16)
    else:
        logger.info_rank0("train_architecture is full")
        _use_orig_params = False

    logger.info_rank0(f"model: {model}")

    if args.train.save_initial_model:
        if args.train.global_rank == 0:
            state_dict = model.state_dict()
            state_dict = {k: v for k, v in state_dict.items() if "lora" in k}
            save_model_weights(args.train.output_dir, model.state_dict(), model_assets=[model_config])

        dist.barrier()
        return

    ops_to_save = convert_ops_to_objects(args.train.ops_to_save)
    model = build_parallelize_model(
        model,
        enable_full_shard=args.train.enable_full_shard,
        enable_mixed_precision=args.train.enable_mixed_precision,
        enable_gradient_checkpointing=args.train.enable_gradient_checkpointing,
        init_device=args.train.init_device,
        enable_fsdp_offload=args.train.enable_fsdp_offload,
        basic_modules=model._no_split_modules,
        enable_reentrant=args.train.enable_reentrant,
        enable_forward_prefetch=args.train.enable_forward_prefetch,
        use_orig_params=_use_orig_params,
        ops_to_save=ops_to_save,
    )

    optimizer = build_optimizer(
        model,
        lr=args.train.lr,
        weight_decay=args.train.weight_decay,
        fused=True,
        optimizer_type=args.train.optimizer,
        param_groups=get_param_groups(model, args.train.lr, args.train.vit_lr),
    )

    if args.train.global_rank == 0:
        if args.train.use_wandb:
            wandb.init(
                project=args.train.wandb_project,
                name=args.train.wandb_name,
                config={**vars(args.model), **vars(args.data), **vars(args.train)},  # flatten dict
            )

        model_assets = [model_config]
        save_model_assets(args.train.model_assets_dir, model_assets)

    if args.train.profile_this_rank:
        profiler = helper.create_profiler(
            start_step=args.train.profile_start_step,
            end_step=args.train.profile_end_step,
            trace_dir=args.train.profile_trace_dir,
            record_shapes=args.train.profile_record_shapes,
            profile_memory=args.train.profile_profile_memory,
            with_stack=args.train.profile_with_stack,
            global_rank=args.train.global_rank,
        )
        profiler.start()

    # Build diffusion scheduler: FlowMatchScheduler
    flow_scheduler = FlowMatchScheduler(
        shift=5,
        sigma_min=0.0,
        extra_one_step=True,
    )

    total_train_steps = args.train.train_steps * args.train.num_train_epochs
    lr_scheduler = build_lr_scheduler(
        optimizer,
        train_steps=total_train_steps,
        lr=args.train.lr,
        lr_min=args.train.lr_min,
        lr_decay_style=args.train.lr_decay_style,
        lr_decay_ratio=args.train.lr_decay_ratio,
        lr_warmup_ratio=args.train.lr_warmup_ratio,
        lr_start=args.train.lr_start,
    )

    start_epoch, start_step, global_step = 0, 0, 0
    save_checkpoint_path = None
    environ_meter = EnvironMeter(
        config=model_config,
        global_batch_size=args.train.global_batch_size,
        empty_cache_steps=args.train.empty_cache_steps,
    )

    if args.train.load_checkpoint_path:
        state = {"model": model, "optimizer": optimizer, "extra_state": {}}  # cannot be None
        Checkpointer.load(args.train.load_checkpoint_path, state)
        global_step = state["extra_state"]["global_step"]
        start_epoch = global_step // args.train.train_steps
        start_step = global_step % args.train.train_steps
        lr_scheduler.load_state_dict(state["extra_state"]["lr_scheduler"])
        train_dataloader.load_state_dict(state["extra_state"]["train_dataloader"])
        environ_meter.load_state_dict(state["extra_state"]["environ_meter"])
        torch.set_rng_state(state["extra_state"]["torch_rng_state"])
        if start_step == 0:  # resume at the end of epoch
            iter(train_dataloader)  # clear resume state and prefetch data

        dist.barrier()
        logger.info_rank0(f"Load distributed checkpoint from {args.train.load_checkpoint_path} successfully!")

    helper.empty_cache()

    model_fwd_context, model_bwd_context = build_activation_offloading_context(
        args.train.enable_activation_offload,
        args.train.enable_gradient_checkpointing,
        args.train.activation_gpu_limit,
    )

    helper.empty_cache()
    model.train()
    logger.info(
        f"rank{args.train.local_rank} Start training, train_steps: {args.train.train_steps}, epochs: {args.train.num_train_epochs}"
    )

    flow_scheduler.set_timesteps(1000, training=True)

    for epoch in range(start_epoch, args.train.num_train_epochs):
        if hasattr(train_dataloader, "set_epoch"):
            train_dataloader.set_epoch(epoch)

        data_loader_tqdm = trange(
            args.train.train_steps,
            desc=f"Epoch {epoch + 1}/{args.train.num_train_epochs}",
            total=args.train.train_steps,
            initial=start_step,
            disable=args.train.local_rank != 0,
        )
        epoch_start_time = time.time()
        data_iterator = iter(train_dataloader)

        epoch_loss = 0
        for _ in range(start_step, args.train.train_steps):
            global_step += 1
            synchronize()
            total_loss = 0
            start_time = time.time()
            try:
                micro_batches: List[Dict[str, Any]] = next(data_iterator)
            except StopIteration:
                logger.info(f"epoch:{epoch} Dataloader finished with drop_last {args.data.drop_last}")
                break

            for micro_batch in micro_batches:
                environ_meter.add(micro_batch, model_type="wan")
                latents = micro_batch["latents"].to(model.device)
                prompt_emb = micro_batch["prompt_emb"]
                if args.train.micro_batch_size > 1:
                    prompt_emb["context"] = prompt_emb["context"].squeeze(1).to(model.device)
                    image_emb = micro_batch["image_emb"]
                    if "clip_feature" in image_emb:
                        image_emb["clip_feature"] = image_emb["clip_feature"].squeeze(1).to(model.device)
                    if "y" in image_emb:
                        image_emb["y"] = image_emb["y"].squeeze(1).to(model.device)
                else:
                    prompt_emb["context"] = prompt_emb["context"][0].to(model.device)
                    image_emb = micro_batch["image_emb"]
                    if "clip_feature" in image_emb:
                        image_emb["clip_feature"] = image_emb["clip_feature"][0].to(model.device)
                    if "y" in image_emb:
                        image_emb["y"] = image_emb["y"][0].to(model.device)

                noise = torch.randn_like(latents)
                timestep_id = torch.randint(0, flow_scheduler.num_train_timesteps, (latents.size(0),))
                timestep = flow_scheduler.timesteps[timestep_id].to(latents.dtype).to(latents.device)
                # noise and target
                noisy_latents = flow_scheduler.add_noise(
                    latents,
                    noise,
                    timestep,
                    args.train.micro_batch_size,
                    args.train.enable_mixed_precision,
                )
                training_target = flow_scheduler.training_target(latents, noise, timestep)
                # predict noise
                with model_fwd_context:
                    noise_pred = model.forward(
                        noisy_latents,
                        timestep=timestep,
                        **prompt_emb,
                        **image_emb,
                    )
                    # MSE loss with weights
                    loss = F.mse_loss(noise_pred.float(), training_target.float(), reduction="none")
                    weight = flow_scheduler.training_weight(timestep, args.train.micro_batch_size)
                    # shape: [B, ...], weight: [B]
                    loss = (loss.view(latents.size(0), -1).mean(dim=1) * weight).mean() / len(micro_batches)

                with model_bwd_context:
                    loss.backward()

                total_loss += loss.item()
                del micro_batch

            grad_norm = veomni_clip_grad_norm(model, args.train.max_grad_norm)

            optimizer.step()
            lr_scheduler.step()
            optimizer.zero_grad()
            if hasattr(grad_norm, "full_tensor"):
                grad_norm = grad_norm.full_tensor().item()

            total_loss, grad_norm = all_reduce((total_loss, grad_norm), group=get_parallel_state().fsdp_group)
            epoch_loss += total_loss
            synchronize()
            delta_time = time.time() - start_time
            lr = max(lr_scheduler.get_last_lr())
            train_metrics = environ_meter.step(delta_time, global_step=global_step)

            data_loader_tqdm.set_postfix_str(
                f"loss: {total_loss:.4f}, grad_norm: {grad_norm:.2f}, lr: {lr:.2e}, step_time: {delta_time:.2f}s"
            )
            data_loader_tqdm.update()

            if args.train.global_rank == 0:
                if args.train.use_wandb:
                    train_metrics.update(
                        {"training/loss": total_loss, "training/grad_norm": grad_norm, "training/lr": lr}
                    )
                    wandb.log(train_metrics, step=global_step)

            if args.train.profile_this_rank and global_step <= args.train.profile_end_step:
                profiler.step()
                if global_step == args.train.profile_end_step:
                    profiler.stop()

            if args.train.save_steps and global_step % args.train.save_steps == 0:
                helper.empty_cache()
                save_checkpoint_path = os.path.join(args.train.save_checkpoint_path, f"global_step_{global_step}")
                state = {
                    "model": model,
                    "optimizer": optimizer,
                    "extra_state": {
                        "global_step": global_step,
                        "lr_scheduler": lr_scheduler.state_dict(),
                        "train_dataloader": train_dataloader.state_dict(),
                        "environ_meter": environ_meter.state_dict(),
                        "torch_rng_state": torch.get_rng_state(),
                    },
                }
                Checkpointer.save(args.train.save_checkpoint_path, state, global_steps=global_step)
                if args.train.global_rank == 0:
                    save_hf_weights(args, save_checkpoint_path, model_assets)

        data_loader_tqdm.close()
        epoch_time = time.time() - epoch_start_time
        start_step = 0
        helper.print_device_mem_info(f"VRAM usage after epoch {epoch + 1}")
        if args.train.global_rank == 0:
            logger.info_rank0(
                f"Epoch {epoch + 1} completed, epoch_time={epoch_time:.4f}s, epoch_loss={epoch_loss / args.train.train_steps:.4f}"
            )
        if args.train.global_rank == 0:
            if args.train.use_wandb:
                wandb.log({"training/loss_per_epoch": epoch_loss / args.train.train_steps}, step=global_step)
        if args.train.save_epochs and (epoch + 1) % args.train.save_epochs == 0:
            helper.empty_cache()
            save_checkpoint_path = os.path.join(args.train.save_checkpoint_path, f"global_step_{global_step}")
            state = {
                "model": model,
                "optimizer": optimizer,
                "extra_state": {
                    "global_step": global_step,
                    "lr_scheduler": lr_scheduler.state_dict(),
                    "train_dataloader": train_dataloader.state_dict(),
                    "environ_meter": environ_meter.state_dict(),
                    "torch_rng_state": torch.get_rng_state(),
                },
            }
            Checkpointer.save(args.train.save_checkpoint_path, state, global_steps=global_step)
            if args.train.global_rank == 0:
                save_hf_weights(args, save_checkpoint_path, model_assets)

    synchronize()
    # release memory
    del optimizer, lr_scheduler
    helper.empty_cache()

    dist.barrier()
    dist.destroy_process_group()


def save_hf_weights(args, save_checkpoint_path, model_assets):
    hf_weights_path = os.path.join(save_checkpoint_path, "hf_ckpt")
    model_state_dict = ckpt_to_state_dict(
        save_checkpoint_path=save_checkpoint_path,
        output_dir=args.train.output_dir,
        ckpt_manager=args.train.ckpt_manager,
    )
    if args.train.train_architecture == "lora":
        model_state_dict = {k: v for k, v in model_state_dict.items() if "lora" in k}
    save_model_weights(
        hf_weights_path,
        model_state_dict,
        model_assets=model_assets,
    )
    logger.info_rank0(f"Huggingface checkpoint saved at {hf_weights_path} successfully!")


if __name__ == "__main__":
    main()



# copied from https://huggingface.co/deepseek-ai/DeepSeek-V3/blob/main/inference/kernel.py
from typing import Tuple

import torch
import triton
import triton.language as tl
from triton import Config


@triton.jit
def act_quant_kernel(x_ptr, y_ptr, s_ptr, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis=0)
    offs = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    x = tl.load(x_ptr + offs).to(tl.float32)
    s = tl.max(tl.abs(x)) / 448.0
    y = x / s
    y = y.to(y_ptr.dtype.element_ty)
    tl.store(y_ptr + offs, y)
    tl.store(s_ptr + pid, s)


def act_quant(x: torch.Tensor, block_size: int = 128) -> Tuple[torch.Tensor, torch.Tensor]:
    assert x.is_contiguous()
    assert x.size(-1) % block_size == 0
    y = torch.empty_like(x, dtype=torch.float8_e4m3fn)
    s = x.new_empty(*x.size()[:-1], x.size(-1) // block_size, dtype=torch.float32)
    grid = lambda meta: (triton.cdiv(x.numel(), meta["BLOCK_SIZE"]),)  # noqa: E731
    act_quant_kernel[grid](x, y, s, BLOCK_SIZE=block_size)
    return y, s


@triton.jit
def weight_dequant_kernel(x_ptr, s_ptr, y_ptr, M, N, BLOCK_SIZE: tl.constexpr):
    pid_m = tl.program_id(axis=0)
    pid_n = tl.program_id(axis=1)
    n = tl.cdiv(N, BLOCK_SIZE)
    offs_m = pid_m * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    offs_n = pid_n * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    offs = offs_m[:, None] * N + offs_n[None, :]
    mask = (offs_m[:, None] < M) & (offs_n[None, :] < N)
    x = tl.load(x_ptr + offs, mask=mask).to(tl.float32)
    s = tl.load(s_ptr + pid_m * n + pid_n)
    y = x * s
    tl.store(y_ptr + offs, y, mask=mask)


def weight_dequant(x: torch.Tensor, s: torch.Tensor, block_size: int = 128) -> torch.Tensor:
    assert x.is_contiguous() and s.is_contiguous()
    assert x.dim() == 2 and s.dim() == 2
    M, N = x.size()
    y = torch.empty_like(x, dtype=torch.get_default_dtype())
    grid = lambda meta: (triton.cdiv(M, meta["BLOCK_SIZE"]), triton.cdiv(N, meta["BLOCK_SIZE"]))  # noqa: E731
    weight_dequant_kernel[grid](x, s, y, M, N, BLOCK_SIZE=block_size)
    return y


fp8_gemm_configs = [
    Config({"BLOCK_SIZE_M": block_m, "BLOCK_SIZE_N": block_n, "BLOCK_SIZE_K": 128}, num_stages=num_stages, num_warps=8)
    for block_m in [16, 32, 64]
    for block_n in [32, 64, 128]
    for num_stages in [3, 4, 5, 6]
]


@triton.autotune(configs=fp8_gemm_configs, key=["N", "K"])
@triton.jit
def fp8_gemm_kernel(
    a_ptr,
    b_ptr,
    c_ptr,
    a_s_ptr,
    b_s_ptr,
    M,
    N: tl.constexpr,
    K: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
):
    pid_m = tl.program_id(axis=0)
    pid_n = tl.program_id(axis=1)
    k = tl.cdiv(K, BLOCK_SIZE_K)
    offs_m = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_n = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + offs_m[:, None] * K + offs_k[None, :]
    b_ptrs = b_ptr + offs_n[None, :] * K + offs_k[:, None]
    a_s_ptrs = a_s_ptr + offs_m * k
    b_s_ptrs = b_s_ptr + (offs_n // BLOCK_SIZE_K) * k

    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for i in range(k):
        a = tl.load(a_ptrs, mask=offs_k[None, :] < K - i * BLOCK_SIZE_K, other=0.0)
        b = tl.load(b_ptrs, mask=offs_k[:, None] < K - i * BLOCK_SIZE_K, other=0.0)
        a_s = tl.load(a_s_ptrs)
        b_s = tl.load(b_s_ptrs)
        accumulator += tl.dot(a, b) * a_s[:, None] * b_s[None, :]
        a_ptrs += BLOCK_SIZE_K
        b_ptrs += BLOCK_SIZE_K
        a_s_ptrs += 1
        b_s_ptrs += 1
    c = accumulator.to(c_ptr.dtype.element_ty)
    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + offs_m[:, None] * N + offs_n[None, :]
    mask = (offs_m[:, None] < M) & (offs_n[None, :] < N)
    tl.store(c_ptrs, c, mask=mask)


def fp8_gemm(a: torch.Tensor, a_s: torch.Tensor, b: torch.Tensor, b_s: torch.Tensor):
    assert a.is_contiguous() and b.is_contiguous()
    assert a_s.is_contiguous() and b_s.is_contiguous()
    K = a.size(-1)
    M = a.numel() // K
    N = b.size(0)
    c = a.new_empty(*a.size()[:-1], N, dtype=torch.get_default_dtype())
    grid = lambda META: (triton.cdiv(M, META["BLOCK_SIZE_M"]), triton.cdiv(N, META["BLOCK_SIZE_N"]))  # noqa: E731
    fp8_gemm_kernel[grid](a, b, c, a_s, b_s, M, N, K)
    return c



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Callable, Dict, List

import torch

from ...utils.import_utils import is_video_audio_available
from ..constants import TYPE2INDEX
from .image_utils import fetch_images
from .preprocess import conv_preprocess


if is_video_audio_available():
    from .audio_utils import fetch_audios
    from .video_utils import fetch_videos
else:

    def fetch_videos(*args, **kwargs):
        return [], []

    def fetch_audios(*args, **kwargs):
        return []


if TYPE_CHECKING:
    from ...models.seed_omni import SeedOmniProcessor
    from .multimodal_chat_template import MultimodalChatTemplate


def mask_before_position_id_func(input_ids: torch.Tensor):
    """Mask special multimodal tokens in input_ids to input_mm_token for position_id.
    Only supports special image tokens now. (input_image_id=-200, output_image_id=-201->-200)
    Similar to veomni.module.seed_omni.modeling_seed_omni.mask_before_text_encoder

    Args:
        input_ids (torch.Tensor)

    Returns:
        input_ids (torch.Tensor)
    """
    for modality in ["image", "video", "audio"]:
        output_mask = input_ids == TYPE2INDEX["output"][modality]
        input_mask = input_ids == TYPE2INDEX["input"][modality]
        input_ids = torch.where(output_mask | input_mask, TYPE2INDEX["input"][modality], input_ids)
    return input_ids


def mask_input_ids(modality_info: Dict, input_ids: torch.Tensor):
    """Mask special multimodal tokens in input_ids to 0 for text_encoder.word_embedding.
    And return masks including: image_input_mask, image_output_mask, etc
    For example:
        input_ids:                  torch.tensor([-200, -200,   2,  -200,   -200,   4,  5,  6,  -201,   -201])
        Returns:
            input_ids:              torch.tensor([0,    0,      2,  0,      0,      4,  5,  6,  0,      0   ])
            image_input_mask:       torch.tensor([1,    1,      0,  1,      1,      0,  0,  0,  0,      0   ])
            image_output_mask:      torch.tensor([0,    0,      0,  0,      0,      0,  0,  0,  1,      1   ])

    Args:
        input_ids (torch.Tensor)

    Returns:
        input_ids (torch.Tensor)
        mask_dict (Dict) : {modal}_[input/output]_mask.
    """
    mask_dict = {}
    for data_type in modality_info.keys():
        for modal in modality_info[data_type]:
            mask = input_ids == TYPE2INDEX[data_type][modal]
            mask_dict[f"{modal}_{data_type}_mask"] = mask
            input_ids = torch.where(mask, 0, input_ids)
    return input_ids, mask_dict


def process_mm_data(
    conversations, images: List[Any], videos: List[Any], video_audios: List[Any], audio_audios: List[Any]
):
    """
    Processes multi-modal conversation data and aligns images, videos, and audio
    with a corresponding output mask indicating whether the data was produced by the assistant.

    Parameters:
    ----------
    conversations : List[List]
    images : List[Any, List of image data in order.
    videos : List[Any], List of video data in order.
    video_audios : List[Any], List of audio tracks corresponding to the videos.
    audio_audios : List[Any], List of standalone audio samples.

    Returns:
    -------
    conv_images : List[Any], List of images in the order they appeared in conversations.
    conv_videos : List[Any], List of videos in the order they appeared in conversations.
    conv_audios : List[Any], List of all audio data, including both video audio and standalone audio.

    mask : Dict[str, torch.BoolTensor]
        A dictionary with modality names as keys ("image", "video", "audio"), and boolean tensors
        indicating whether each sample was produced by the assistant (True) or the user (False).

    Example:
    --------
    Input:
        conversations = [
            ["user", ["video"], ["audio"], ["video"], ["text"]],
            ["assistant", ["audio"]]
        ]
        videos = ["video1", "video2"]
        video_audios = ["v_audio1", "v_audio2"]
        audio_audios = ["audio1", "audio2"]

    Output:
        conv_videos = ["video1", "video2"]
        conv_audios = ["v_audio1", "audio1", "v_audio2", "audio2"]
        mask["video"] = tensor([False, False])              # user videos
        mask["audio"] = tensor([False, False, False, True]) # user+assistant audios
    """
    images, videos, video_audios, audio_audios = iter(images), iter(videos), iter(video_audios), iter(audio_audios)
    conv_images, conv_videos, conv_audios = [], [], []
    mask = defaultdict(list)
    for conversation in conversations:
        role = conversation[0]
        is_output = role == "assistant"
        for message in conversation[1:]:
            data_type = message[0]
            if data_type == "text":
                continue
            elif data_type == "image":
                conv_images.append(next(images))
                mask["image"].append(is_output)
            elif data_type == "video":
                conv_videos.append(next(videos))
                conv_audios.append(next(video_audios))
                mask["video"].append(is_output)
                mask["audio"].append(is_output)
            elif data_type == "audio":
                conv_audios.append(next(audio_audios))
                mask["audio"].append(is_output)
            else:
                raise ValueError(f"Unknown data type: {data_type}")
    mask = {key: torch.tensor(value).type(torch.bool) for key, value in mask.items()}
    return conv_images, conv_videos, conv_audios, mask


def get_multimodal_configs(modality_input: Dict, multimodal_output_mask: Dict):
    multimodal_configs, config_repr = {}, {}
    for key in modality_input.keys():
        config_key = key.split("_", 2)[-1]
        if config_key != "features":
            config_repr[config_key] = modality_input[key]
    for config_key, repr in config_repr.items():
        multimodal_configs[config_key] = {}
        for modal, mm_mask in multimodal_output_mask.items():
            if (
                f"{modal}_input_{config_key}" not in modality_input
                and f"{modal}_output_{config_key}" not in modality_input
            ):
                continue
            input_config = modality_input.get(f"{modal}_input_{config_key}", torch.empty_like(repr))
            output_config = modality_input.get(f"{modal}_output_{config_key}", torch.empty_like(repr))

            config = torch.zeros_like(repr)
            config = config.repeat_interleave(mm_mask.shape[0], dim=0)

            config[mm_mask] = output_config
            config[~mm_mask] = input_config

            multimodal_configs[config_key][modal] = config
    return multimodal_configs


def keep_input_only(multimodal_config: Dict, multimodal_output_mask: Dict):
    """Only keep the input data in multimodal_config. Used when use_special_rope=False.
    When use_special_rope=False, only do special_rope on input_multimodal_data.
    For example: 2d_rope on input_image_token, but 1d_rope on output_image_token.
    """
    for config in multimodal_config.keys():
        for modal in multimodal_config[config].keys():
            multimodal_config[config][modal] = multimodal_config[config][modal][~multimodal_output_mask[modal]]


def encode_multimodal_sample(
    sample: Dict[str, Any],
    processor: "SeedOmniProcessor",
    chat_template: "MultimodalChatTemplate",
    position_id_func: "Callable",
    modality_info: Dict,
    use_special_rope=False,  # 2d rope position id for image generation
    **kwargs,
) -> Dict[str, List[int]]:
    model_inputs = {}
    source = sample.pop("source_name") if "source_name" in sample else kwargs["source_name"]
    modality = set(modality_info["input"] + modality_info["output"])
    conversations = sample["text"] if source == "fineweb_100BT" else sample["conversations"]  # text-only data
    if isinstance(conversations, bytes):
        conversations = json.loads(conversations.decode("utf-8"))
    conversations = conv_preprocess(source, conversations, **kwargs)
    processor_input = {}

    if "image" in modality:
        images = fetch_images(sample.get("images", []), **kwargs)
    else:
        images = []
    if "video" in modality:
        videos, video_audios = fetch_videos(sample.get("videos", []), **kwargs)
        if "audio" not in modality:
            video_audios = [None] * len(videos)
    else:
        videos, video_audios = [], []
    if "audio" in modality:
        audio_audios = fetch_audios(sample.get("audios", []), **kwargs)
    else:
        audio_audios = []

    images, videos, audios, multimodal_output_mask = process_mm_data(
        conversations, images, videos, video_audios, audio_audios
    )

    if images:
        processor_input.update(
            {
                "input_images": [img for img, mask in zip(images, multimodal_output_mask["image"]) if not mask],
                "output_images": [img for img, mask in zip(images, multimodal_output_mask["image"]) if mask],
            }
        )
    if videos:
        processor_input.update(
            {
                "input_videos": [vid for vid, mask in zip(videos, multimodal_output_mask["video"]) if not mask],
                "output_videos": [img for img, mask in zip(videos, multimodal_output_mask["video"]) if mask],
            }
        )
    if audios and "audio" in modality:
        processor_input.update(
            {
                "input_audios": [aud for aud, mask in zip(audios, multimodal_output_mask["audio"]) if not mask],
                "output_audios": [aud for aud, mask in zip(audios, multimodal_output_mask["audio"]) if mask],
            }
        )

    modality_input = processor(return_tensors="pt", **processor_input)
    multimodal_config = get_multimodal_configs(modality_input, multimodal_output_mask)
    text_inputs = chat_template.encode_messages(conversations, **multimodal_config)
    model_inputs.update(modality_input)
    model_inputs.update(text_inputs)

    # position_ids (dim, len)
    if position_id_func is None:  # default position_ids
        position_ids = torch.arange(0, len(text_inputs["input_ids"])).unsqueeze(0)
    else:  # customized position_ids
        input_ids = text_inputs["input_ids"].clone()
        attention_mask = text_inputs["attention_mask"].clone()
        if use_special_rope:
            input_ids = mask_before_position_id_func(input_ids)
        else:
            keep_input_only(multimodal_config, multimodal_output_mask)
        position_ids = position_id_func(
            input_ids=input_ids.unsqueeze(0), attention_mask=attention_mask.unsqueeze(0), **multimodal_config
        )["position_ids"]
    model_inputs["position_ids"] = position_ids

    input_ids, mask_dict = mask_input_ids(modality_info, model_inputs["input_ids"])
    model_inputs["input_ids"] = input_ids
    model_inputs.update(mask_dict)
    return [model_inputs]


def encode_multimodal_sample_inference(
    sample: Dict[str, Any],
    processor: "SeedOmniProcessor",
    chat_template: "MultimodalChatTemplate",
    position_id_func: "Callable",
    modality_info: Dict,
    force_image_gen: bool,
    **kwargs,
):
    model_inputs = {}
    modality = set(modality_info["input"] + modality_info["output"])
    conversations = sample["conversations"]

    processor_input = {}
    if "image" in modality:
        images = fetch_images(sample.get("images", []), **kwargs)
    else:
        images = []
    if "video" in modality:
        videos, video_audios = fetch_videos(sample.get("videos", []), **kwargs)
        if "audio" not in modality:
            video_audios = [None] * len(videos)
    else:
        videos, video_audios = [], []
    if "audio" in modality:
        audio_audios = fetch_audios(sample.get("audios", []), **kwargs)
    else:
        audio_audios = []

    images, videos, audios, multimodal_output_mask = process_mm_data(
        conversations, images, videos, video_audios, audio_audios
    )

    if images:
        processor_input["input_images"] = images
    if videos:
        processor_input["input_videos"] = videos
    if audios and "audio" in modality:
        processor_input["input_audios"] = audios

    modality_input = processor(return_tensors="pt", **processor_input)
    multimodal_config = get_multimodal_configs(modality_input, multimodal_output_mask)
    text_inputs = chat_template.encode_messages(conversations, **multimodal_config)

    if force_image_gen:
        text_inputs["input_ids"] = torch.cat(
            [text_inputs["input_ids"], torch.tensor([chat_template.image_start_id])],
            dim=-1,
        )
        text_inputs["attention_mask"] = torch.cat([text_inputs["attention_mask"], torch.tensor([1])], dim=-1)

    model_inputs.update(modality_input)
    model_inputs.update(text_inputs)

    # position_ids (dim, len)
    if position_id_func is None:  # default position_ids
        position_id_returns = {"position_ids": torch.arange(0, len(text_inputs["input_ids"])).unsqueeze(0)}
    else:  # customized position_ids
        input_ids = text_inputs["input_ids"].clone()
        attention_mask = text_inputs["attention_mask"].clone()
        position_id_returns = position_id_func(
            input_ids=input_ids.unsqueeze(0), attention_mask=attention_mask.unsqueeze(0), **multimodal_config
        )

    model_inputs.update(position_id_returns)

    input_ids, mask_dict = mask_input_ids(modality_info, model_inputs["input_ids"])
    model_inputs["input_ids"] = input_ids
    model_inputs.update(mask_dict)
    return [model_inputs]



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import transformers

from . import (
    janus_foundation,
    llama_foundation,
    qwen2_5_omni_foundation,
    qwen2_5_vl_foundation,
    qwen2_vl_foundation,
    qwen3_moe_foundation,
)
from .base import BaseFoundationConfigMixin, BaseFoundationModelMixin


__all__ = [
    "qwen2_vl_foundation",
    "janus_foundation",
    "llama_foundation",
    "qwen2_5_vl_foundation",
    "qwen2_5_omni_foundation",
    "BaseFoundationModelMixin",
    "BaseFoundationConfigMixin",
    "qwen3_moe_foundation",
]

if transformers.__version__ >= "4.56.0":
    from . import seed_oss_foundation

    __all__ += ["seed_oss_foundation"]



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Adapted from https://github.com/sgl-project/sglang/blob/main/python/sglang/srt/model_loader/loader.py

from abc import ABC

import torch
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForCausalLM,
    AutoModelForImageTextToText,
    AutoModelForVision2Seq,
    AutoProcessor,
    PretrainedConfig,
    PreTrainedModel,
)
from transformers.modeling_utils import no_init_weights

from ..utils import logging
from ..utils.registry import Registry
from .module_utils import init_empty_weights, load_model_weights


MODELING_REGISTRY = Registry("Modeling")
MODEL_CONFIG_REGISTRY = Registry("ModelConfig")
MODEL_PROCESSOR_REGISTRY = Registry("ModelProcessor")

logger = logging.get_logger(__name__)


def get_model_config(config_path: str, force_use_huggingface: bool = False, **kwargs):
    if force_use_huggingface:
        logger.info_rank0("[CONFIG] Force loading model config from Huggingface.")
        return AutoConfig.from_pretrained(config_path, **kwargs)
    else:
        try:  # first load from hf, then replace with veomni
            config = AutoConfig.from_pretrained(config_path, **kwargs)
            model_type = config.model_type
            if model_type in MODEL_CONFIG_REGISTRY.valid_keys():
                kwargs.pop("trust_remote_code", None)
                config = MODEL_CONFIG_REGISTRY[model_type]().from_pretrained(config_path, **kwargs)
                logger.info_rank0(
                    f"[CONFIG] Loading {model_type} from Huggingface and replaced with customized config."
                )
                return config
            else:
                logger.info_rank0(
                    f"[CONFIG] Loading {model_type} from Huggingface as no customized config registered."
                )
                return config
        except Exception:  # load from veomni
            config_dict, _ = PretrainedConfig.get_config_dict(config_path, **kwargs)
            model_type = config_dict["model_type"]
            logger.info_rank0(f"[CONFIG] Loading {model_type} from custom config.")
            kwargs.pop("trust_remote_code", None)
            return MODEL_CONFIG_REGISTRY[model_type]().from_pretrained(config_path, **kwargs)


def get_model_processor(processor_path: str, force_use_huggingface: bool = False, **kwargs):
    if force_use_huggingface:
        logger.info_rank0("[PROCESSOR] Force loading model processor from Huggingface.")
        return AutoProcessor.from_pretrained(processor_path, **kwargs)
    else:
        try:  # first load from hf, then replace with veomni
            processor = AutoProcessor.from_pretrained(processor_path, **kwargs)
            processor_class_name = getattr(type(processor), "__name__", None)
            if processor_class_name in MODEL_PROCESSOR_REGISTRY.valid_keys():
                kwargs.pop("trust_remote_code", None)
                processor = MODEL_PROCESSOR_REGISTRY[processor_class_name]().from_pretrained(processor_path, **kwargs)
                logger.info_rank0(
                    f"[PROCESSOR] Loading {processor_class_name} from Huggingface and replaced with customized processor."
                )
                return processor
            else:
                logger.info_rank0(
                    f"[PROCESSOR] Loading {processor_class_name} from Huggingface as no customized processor registered."
                )
                return processor
        except Exception:  # load from veomni
            from transformers.processing_utils import ProcessorMixin
            from transformers.utils import PROCESSOR_NAME, cached_file

            processor_config_file = cached_file(processor_path, PROCESSOR_NAME)
            config_dict, _ = ProcessorMixin.get_processor_dict(processor_config_file, **kwargs)
            processor_class_name = config_dict["processor_class"]
            logger.info_rank0(f"[PROCESSOR] Loading {processor_class_name} from custom processor.")
            kwargs.pop("trust_remote_code", None)
            return MODEL_PROCESSOR_REGISTRY[processor_class_name]().from_pretrained(processor_path, **kwargs)


def get_model_class(model_config: PretrainedConfig, force_use_huggingface: bool = False):
    def get_model_arch_from_config(model_config):
        arch_name = model_config.architectures
        if isinstance(arch_name, list):
            arch_name = arch_name[0]
        return arch_name

    arch_name = get_model_arch_from_config(model_config)
    model_type = model_config.model_type
    if type(model_config) in AutoModelForImageTextToText._model_mapping.keys():  # assume built-in models
        load_class = AutoModelForImageTextToText
    elif type(model_config) in AutoModelForVision2Seq._model_mapping.keys():  # assume built-in models
        load_class = AutoModelForVision2Seq
    elif (
        arch_name is not None
        and "ForCausalLM" in arch_name
        and type(model_config) in AutoModelForCausalLM._model_mapping.keys()
    ):
        load_class = AutoModelForCausalLM
    else:
        load_class = AutoModel
    if force_use_huggingface:
        return load_class
    return MODELING_REGISTRY[model_type](arch_name)


class BaseModelLoader(ABC):
    def __init__(self):
        pass

    def load_model(self, model_config, **kwargs):
        raise NotImplementedError


class HuggingfaceLoader(BaseModelLoader):
    def __init__(self, model_cls: PreTrainedModel):
        super().__init__()
        self.model_cls = model_cls

    def load_model(self, init_kwargs: dict, **kwargs):
        init_kwargs.pop("trust_remote_code", True)

        init_device = kwargs.pop("init_device", "cuda")
        weights_path = kwargs.pop("weights_path", None)
        empty_init = kwargs.pop("empty_init", False)

        logger.info_rank0(
            f"Loading model from Huggingface modeling.\n"
            f"init_device: {init_device}\n"
            f"empty_init: {empty_init}\n"
            f"weights_path: {weights_path}"
        )

        if weights_path is None:  # init empty model from config
            if init_device == "meta":
                with init_empty_weights():
                    logger.info_rank0("Init empty model on meta device from config without init_weights.")
                    model = self.model_cls.from_config(**init_kwargs)
            else:
                with torch.device(init_device):
                    logger.info_rank0("Init empty model from config.")
                    model = self.model_cls.from_config(**init_kwargs)
        else:
            with init_empty_weights():
                model = self.model_cls.from_config(**init_kwargs)
            if not empty_init:
                load_model_weights(model, weights_path, init_device)

        return model


class CustomizedModelingLoader(BaseModelLoader):
    def __init__(self, model_cls: PreTrainedModel):
        super().__init__()
        self.model_cls = model_cls

    def load_model(self, init_kwargs: dict, **kwargs):
        init_kwargs.pop("trust_remote_code", True)

        init_device = kwargs.pop("init_device", "cuda")
        weights_path = kwargs.pop("weights_path", None)
        empty_init = kwargs.pop("empty_init", False)

        logger.info_rank0(
            f"Loading model from customized modeling.\n"
            f"init_device: {init_device}\n"
            f"empty_init: {empty_init}\n"
            f"weights_path: {weights_path}"
        )

        if weights_path is None:  # init empty model from config
            if init_device == "meta":
                with init_empty_weights():
                    logger.info_rank0("Init empty model on meta device from config without init_weights.")
                    model = self.model_cls._from_config(**init_kwargs)
            else:
                with torch.device(init_device):
                    logger.info_rank0("Init empty model from config.")
                    model = self.model_cls._from_config(**init_kwargs)
        else:
            with init_empty_weights(), no_init_weights():
                model = self.model_cls._from_config(**init_kwargs)

            if not empty_init:
                load_model_weights(model, weights_path, init_device)

            # we should tie embeddings after loading weights because init_empty_weights() leads to untied weights,
            if getattr(model.config, "tie_word_embeddings", True):
                try:
                    input_embeddings = model.get_input_embeddings()
                    output_embeddings = model.get_output_embeddings()
                    output_embeddings._parameters["weight"] = input_embeddings._parameters["weight"]
                except Exception as e:
                    logger.info_rank0(f"Failed to tie embeddings: {e}")

        return model


def get_loader(model_config, force_use_huggingface):
    model_cls = get_model_class(model_config, force_use_huggingface)
    if force_use_huggingface:
        loader = HuggingfaceLoader(model_cls=model_cls)
    else:
        loader = CustomizedModelingLoader(model_cls=model_cls)

    # TODO: there's no difference between HuggingfaceLoader and CustomizedModelingLoader except tie_word_embedding. Check if able to merge
    return loader



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ...loader import MODEL_CONFIG_REGISTRY, MODEL_PROCESSOR_REGISTRY, MODELING_REGISTRY


@MODEL_CONFIG_REGISTRY.register("qwen2_5_omni")
def register_qwen2_5_omni_config():
    from .configuration_qwen2_5_omni import Qwen2_5OmniConfig

    return Qwen2_5OmniConfig


@MODELING_REGISTRY.register("qwen2_5_omni")
def register_qwen2_5_omni_modeling(architecture: str):
    from .modeling_qwen2_5_omni import (
        Qwen2_5OmniForConditionalGeneration,
        Qwen2_5OmniTalkerForConditionalGeneration,
        Qwen2_5OmniTalkerModel,
        Qwen2_5OmniThinkerForConditionalGeneration,
        Qwen2_5OmniThinkerTextModel,
        Qwen2_5OmniToken2WavBigVGANModel,
        Qwen2_5OmniToken2WavDiTModel,
        Qwen2_5OmniToken2WavModel,
    )

    if "ForConditionalGeneration" in architecture:
        return Qwen2_5OmniForConditionalGeneration
    if "ThinkerTextModel" in architecture:
        return Qwen2_5OmniThinkerTextModel
    if "ThinkerForConditionalGeneration" in architecture:
        return Qwen2_5OmniThinkerForConditionalGeneration
    if "TalkerModel" in architecture:
        return Qwen2_5OmniTalkerModel
    if "TalkerForConditionalGeneration" in architecture:
        return Qwen2_5OmniTalkerForConditionalGeneration
    if "Token2WavDiTModel" in architecture:
        return Qwen2_5OmniToken2WavDiTModel
    if "Token2WavBigVGANModel" in architecture:
        return Qwen2_5OmniToken2WavBigVGANModel
    if "Token2WavModel" in architecture:
        return Qwen2_5OmniToken2WavModel
    return Qwen2_5OmniForConditionalGeneration


@MODEL_PROCESSOR_REGISTRY.register("Qwen2_5OmniProcessor")
def register_qwen2_5_omni_processor():
    from .processing_qwen2_5_omni import Qwen2_5OmniProcessor

    return Qwen2_5OmniProcessor
